# -*- coding: utf-8 -*-
"""
Created on Thu Dec  9 20:46:09 2021

@author: joverbee
"""
import numpy as np
import logging
from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape
from pyEELSMODEL.core.spectrum import Spectrum, Spectrumshape


logger = logging.getLogger(__name__)


class Deconvolution(Operator):
    """
    Parten class for the deconvolution methods.
    """
    def __init__(self, spectrum, llspectrum):
        """
        Parameters
        ----------
        spectrum: Spectrum or MultiSpectrum
            The spectrum which needs to be deconvolved. Best practice is to
            remove the background before deconvolving.
        llspectrum: Spectrum or MultiSpectrum
            The low loss which is used for the deconvolution


        """
        self.spectrum = spectrum
        self.llspectrum = llspectrum

        if isinstance(spectrum, MultiSpectrum):
            self.spectrum.multidata = self.spectrum.multidata.astype(float)

        self.spectrum.data = self.spectrum.data.astype(float)
        self.llspectrum.data = self.llspectrum.data.astype(float)

        if self.spectrum.size > self.llspectrum.size:
            print('low loss spectrum has to have same size as model')
            print('the low loss will be zero padded to have the same size')
            llspectrum = self.padding(spectrum.get_spectrumshape(), llspectrum)
            self.llspectrum = llspectrum

    def padding(self, specshape, llspectrum):
        """
        Zero padds the low loss if the size is smaller than the core-loss

        Parameters
        ----------
        specshape: Spectrumshape
            The spectrum shape of the core-loss
        llspectrum: Spectrum or MultiSpectrum
            The low loss which is used for the deconvolution

        """

        size_dif = specshape.size - llspectrum.size
        before = size_dif//2
        after = size_dif - before
        print('padding is done')
        print(llspectrum)
        if type(llspectrum) is Spectrum:
            print('low loss is spectrum')
            pad_data = np.pad(llspectrum.data, pad_width=(before, after))
            noffset = llspectrum.offset - llspectrum.dispersion * before
            sph = Spectrumshape(llspectrum.dispersion, noffset, pad_data.size)
            s = Spectrum(sph, data=pad_data)

        elif type(llspectrum) is MultiSpectrum:
            print('low loss is multispectrum')
            pad_w = ((0, 0), (0, 0), (before, after))
            pad_data = np.pad(llspectrum.multidata, pad_width=pad_w)
            noffset = llspectrum.offset - llspectrum.dispersion * before
            sph = MultiSpectrumshape(llspectrum.dispersion,
                                     noffset,
                                     pad_data.shape[-1],
                                     llspectrum.xsize,
                                     llspectrum.ysize)
            s = MultiSpectrum(sph, data=pad_data)

        return s



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 29 16:25:55 2021

@author: joverbee
"""

import sys
from pyEELSMODEL.components.CLedge.hydrogen_coreloss_edge import \
    HydrogenicCoreLossEdge
from pyEELSMODEL.core.spectrum import Spectrumshape
import pytest
import numpy as np
sys.path.append("..")  # Adds higher directory to python modules path.


def test_set_edge():
    specshape = Spectrumshape(1, 100, 1024)
    OK = HydrogenicCoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3,
                                beta=20e-3, element='O', edge='K')
    assert OK.edge == 'K'

    TiL = HydrogenicCoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3,
                                 beta=20e-3, element='Ti', edge='L')
    assert TiL.edge == 'L'

    with pytest.raises(ValueError):
        HydrogenicCoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3,
                               beta=20e-3, element='Ti', edge='M')
    with pytest.raises(ValueError):
        HydrogenicCoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3,
                               beta=20e-3, element='C', edge='L')
    with pytest.raises(ValueError):
        HydrogenicCoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3,
                               beta=20e-3, element='Ag', edge='L')


def test_set_onset_energy():
    specshape = Spectrumshape(1, 100, 1024)
    OK = HydrogenicCoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3,
                                beta=20e-3, element='O', edge='K')
    assert OK.onset_energy == 532.
    CaL = HydrogenicCoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3,
                                 beta=20e-3, element='Ca', edge='L')
    assert CaL.onset_energy == 347.


def test_calculate():
    specshape = Spectrumshape(1, 100, 1024)
    OK = HydrogenicCoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3,
                                beta=20e-3, element='O', edge='K')
    OK.calculate()
    # give some interval to have a least the right order of magnitude right
    # assert OK.data.sum() >= 1e-22
    # assert OK.data.sum() <= 1e-21

    # check if by changing parameter A new data is calculated
    prev_data = np.copy(OK.data)
    A = OK.getparameter(0)
    A.setvalue(1e5)
    OK.calculate()
    np.testing.assert_allclose(OK.data*1e-2, prev_data)


def test_gradient():
    specshape = Spectrumshape(1, 100, 1024)
    OK = HydrogenicCoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3,
                                beta=20e-3, element='O', edge='K')
    OK.calculate()
    OK.getgradient(OK.getparameter(0))
    assert (OK.gradient[0]*1e3 == OK.data).all()


def main():
    test_set_edge()
    test_set_onset_energy()
    test_calculate()
    test_gradient()


if __name__ == "__main__":
    main()



from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge
from pyEELSMODEL.components.CLedge.kohl_coreloss_edge import KohlLossEdge
from pyEELSMODEL.database.Segger_Guzzinati_Kohl.download import download_file
import numpy as np
import logging
import h5py
import os
from pyEELSMODEL import __file__

logger = logging.getLogger(__name__)


class KohlLossEdgeCombined(CoreLossEdge):
    # core loss edge with L3,L2,L1 combined in one edge w fixed ratios
    """
    Calculates the coreloss edges for a group of edges using the GOS from Kohl.
    For instance, the L edge calculates the L3, L2 and L1 edges and puts them
     together with the appropriate prefactors.

    """

    def __init__(self, specshape, A, E0, alpha, beta, element, edge, eshift=0,
                 q_steps=100, dir_path=None, fast=False):
        """

         Parameters
         ----------
         specshape : Spectrumshape
         The spectrum shape used to model
         A : float
         Amplitude of the edge

         E0: float [V]
         The acceleration voltage of the microscope

         alpha: float [rad]
         The convergence angle of the incoming probe

         beta: float [rad]
         The collection angle

         element: string
         The name of the element from which to calculate the edge model.

         edge: string
         The type of edge. (K, L or M)

         eshift: float [eV]
         The shift of the onset energy with respect to the literature value.
         (default: 0)

         q_steps: int
         The number of q points taken into account for the integration over
         the momentum space. The larger the number of q_steps the more
         accurate the calculation. (default: 100)

         dir_path: string
             The filepath indicating where the GOS tables can be found.
             If None, the default path is used.

        fast: bool
        Use vectorized operations for the crossection calculation.

         Returns
         -------
         """
        if dir_path is None:
            dir_ = "/../pyEELSMODEL/database/Segger_Guzzinati_Kohl/"
            self.dir_path = os.path.dirname(
                os.path.dirname(__file__) + dir_
            )
        else:
            self.dir_path = dir_path
        self.file = os.path.join(self.dir_path, 'Segger_Guzzinati_Kohl_1.5.0.gosh')
        if not os.path.exists(self.file):
            download_file(filename=self.file)
        self.onset_path = \
            os.path.dirname(os.path.dirname(__file__) + "/../pyEELSMODEL/")
        self.xsectionlist = []
        max_edge = self.check_maximum_edge(element, edge)

        if edge == 'K':
            xsectionK = KohlLossEdge(specshape, A, E0, alpha, beta, element,
                                     'K1', eshift=eshift, q_steps=q_steps,
                                     dir_path=self.dir_path,fast = fast)

            self.xsectionlist.append(xsectionK)
            super().__init__(specshape, A, E0, alpha, beta, element, 'K1',
                             q_steps=q_steps)

            name = element + ' K edge: ' + str(self.onset_energy) + ' eV'
            self.setdisplayname(name)

        elif (edge == 'L') | (edge == 'M') | (edge == 'N') | (edge == 'O'):
            start_edge = edge + str(max_edge)
            xsection_op = KohlLossEdge(specshape, A, E0, alpha, beta, element,
                                       start_edge,
                                       eshift=eshift, q_steps=q_steps,
                                       dir_path=self.dir_path,fast = fast)
            self.xsectionlist.append(xsection_op)
            for i in range(max_edge - 1):
                try:
                    next_edge = edge + str(max_edge - i - 1)
                    xsection = KohlLossEdge(specshape, A, E0, alpha, beta,
                                            element, next_edge,
                                            eshift=eshift, q_steps=q_steps,
                                            dir_path=self.dir_path,fast = fast)
                    xsection.parameters[0].couple(xsection_op.parameters[0])
                    xsection.parameters[1].couple(xsection_op.parameters[1])
                    xsection.parameters[2].couple(xsection_op.parameters[2])
                    xsection.parameters[3].couple(xsection_op.parameters[3])
                    self.xsectionlist.append(xsection)
                    print(next_edge + ' is used')

                except Exception:
                    print(next_edge + ' is NOT implemented')

            # self.xsectionlist.append(xsection)
            super().__init__(specshape, A, E0, alpha, beta, element,
                             start_edge, eshift=eshift, q_steps=100)
            name = element + edge + ' edge: ' + str(self.onset_energy) + ' eV'
            self.setdisplayname(name)

        self.manageparameters()
        self.calculate()

    def check_maximum_edge(self, element, edge):
        """
        Check which is the lowest energy loss edge (the highest number)
        available. Some elements only have a L1 but do not have a L3 edge.
        This function can tell which number is the highest for the given
        element and edge.

        Parameters
        ----------
        element : string
            The element from which the information should be retrieved.
        edge : string
            The edge from which the information should be retrieved.

        Returns
        -------
        max_value: uint > 0
            The integer from the edge having the lowest energy onset.

        """

        file = os.path.join(self.onset_path, 'element_info.hdf5')
        with h5py.File(file, 'r') as f:
            edges = list(f[element].keys())

        sub_edges = []
        for char in edges:
            if edge in char:
                sub_edges.append(int(char[-1]))

        max_value = max(sub_edges)
        return max_value

    def manageparameters(self):
        # erase existing parameter list and replace with longer list
        self.parameters = []
        self.gradient = []
        for xsection in self.xsectionlist:
            for par in xsection.parameters:
                self._addparameter(par)

    def set_onset_energy(self):
        """
        Sets the onset energy of the grouped edge.
        """
        self.onset_energy = self.xsectionlist[
            0].onset_energy  # first part of cross section determines onset

    def set_edge(self, edge):
        self.edge = self.xsectionlist[0].edge

    def calculate_cross_section(self):
        """
        Calculates the cross section of the combined edge
        """
        cross_section = np.zeros(self.size)
        for xsection in self.xsectionlist:
            if xsection.suppress==False:
                cross_section += xsection.calculate_cross_section()
            else:
                print('subcomponent is suppressed, to unset call setsuppress(False) on the specific subcomponent:',xsection.name)
        return cross_section

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        
        pA = self.parameters[0]
        p2 = self.parameters[1]
        p3 = self.parameters[2]
        p4 = self.parameters[3]

        if p2.ischanged() or p3.ischanged() or p4.ischanged():
            self.cross_section = self.calculate_cross_section()
            self.data = pA.getvalue() * self.cross_section

        if pA.ischanged():
            self.data = pA.getvalue() * self.cross_section
        self.setunchanged()



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy import interpolate
from tqdm import tqdm
from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape
from pyEELSMODEL.misc.pdf_maker import PDF
import logging
logger = logging.getLogger(__name__)


class Align(Operator):
    """
    This class is the super class of different types of alignment procedures.
    For instance the zero loss alignment using the maximum value or model based
    fitting with Gaussian/Lorentzian. Also other methods of alignment such as
    the cross correlation can be used as child class
    """
    def __init__(self, multispectrum, other_spectra, cropping,
                 signal_range=None, zero_index=None):
        self.multispectrum = multispectrum

        if other_spectra is None:
            self.other_spectra = []
        else:
            if self.check_valid_other_spectra(other_spectra):
                self.other_spectra = other_spectra
        self.update_multispectra()
        self.cropping = cropping
        self.signal_range = signal_range
        self.zero_index = zero_index
        self.shift = None
        self.index_shift = None
        self.aligned = None
        self.aligned_others = []
        self.method = None
        self.pdf_path = r'E:\eelsmodel_pdfs'  # directory to put the pdfs

    @property
    def shift(self):
        """
        The determined shift using a method which is given in the sub-class

        Returns
        -------
        self._shift: 2d numpy array
        """
        return self._shift

    @shift.setter
    def shift(self, shift):
        """
        Sets the shift
        Parameters
        ----------
        shift : 2d numpy array
            The shift calculated using the different methods

        Returns
        -------
        None.
        """

        self._shift = shift

    @property
    def index_shift(self):
        """
        The index shift, this will be used to align the spectra using the
        roll function

        Returns
        -------
        self._index_shift: 2d numpy array (int)
            The index shift.
        """
        return self._index_shift

    @index_shift.setter
    def index_shift(self, shift):
        """
        Sets the index shift

        Parameters
        ----------
        shift : 2d numpy array (int)
            The index shift calculated using the different methods

        Returns
        -------
        None.
        """
        self._index_shift = shift

    @property
    def zero_index(self):
        """
        The reference index for the alignment of the fast align.

        Returns
        -------
        self._zero_index: int
        """
        return self._zero_index

    @zero_index.setter
    def zero_index(self, E):
        """
        Sets the reference index with the used energy E (eV).

        Returns
        -------
        None
        """
        if E is None:
            self._zero_index = None
            return

        E0 = self.multispectrum.energy_axis[0]
        E1 = self.multispectrum.energy_axis[-1]

        if (E >= E1) or (E <= E0):

            raise ValueError(r'The given energy value is not inside the energy'
                             r' axis so it cannot, align on this energy. '
                             r'Please take a energy value inside the axis.')
        self._zero_index = self.multispectrum.get_energy_index(E)

    @property
    def signal_range(self):
        """
        The signal range used in the determination of the shift/index_shift

        Returns
         -------
        self._signal_range: tuple
        """
        return self._signal_range

    @signal_range.setter
    def signal_range(self, signal_range):
        """
        Sets the signal range in eV

        Parameters
        ----------
        signal_range : tuple
            A tuple containing the signal range which is used for the
            alignment. The energy range is given in eV.

        Returns
        -------
        None
        """

        if signal_range is None:
            signal_range = (self.multispectrum.energy_axis[0],
                            self.multispectrum.energy_axis[-1])

        if signal_range[1] < signal_range[0]:
            raise ValueError(r'Second energy in the signal range should be '
                             r'bigger then the first one')

        if signal_range[0] < self.multispectrum.energy_axis[0]:
            signal_range[0] = self.multispectrum.energy_axis[0]
            logger.warning(r'Begin energy is set to start of energy axis')

        if signal_range[1] > self.multispectrum.energy_axis[-1]:
            signal_range[1] = self.multispectrum.energy_axis[-1]
            logger.warning(r'End energy is set to end of energy axis')

        self._signal_range = signal_range

    def check_valid_other_spectra(self, other_spectra):
        """
        Checks whether the given other_spectra have the same x and y size
        which is necessary to have in order to correct the drift.

        Parameters
        ----------
        other_spectra : list of MultiSpectrum
            A list containing the multispectra which uses the same alignment.

        Returns
        -------
        bool: True
            Returns True if no error is thrown.

        """
        for spec in other_spectra:
            if (spec.xsize != self.multispectrum.xsize) or \
                    (spec.ysize != self.multispectrum.ysize):
                raise ValueError(r'Scan size of the other spectra is not '
                                 r'equal')
                return False

            # elif (spec.size != self.multispectrum.size) or \
            #         (spec.dispersion != self.multispectrum.dispersion):
            #     raise ValueError(r'Dispersion or offset is not the same for '
            #                      r'the other spectra')

        return True

    def update_multispectra(self):
        """
        If the zero loss is not centered around the zero energy then we cut
        too much of when cropping. This is resolved by doing a first quick
        update on the offset energy of the multispectrum by taking the median
        value of the energy at maximum intensity.
        """
        co = np.argmax(self.multispectrum.multidata[:, :], axis=2)
        zlp_pos = self.multispectrum.energy_axis[co]
        zlp_med = np.median(zlp_pos)

        self.multispectrum.offset -= zlp_med

        for spec in self.other_spectra:
            spec.offset -= zlp_med

    def fast_align(self):
        """
        Aligns the data using a roll without any interpolation.
        parameter. The new aligned spectrum is saved in the aligned attribute.
        The additional aligned spectra can be found in the aligned_others
        attribute.

        Returns
        -------
        None.
        """
        if self.index_shift is None:
            print('The shift value should first be calculated before applying'
                  ' it')
            return None

        # make the aligned other empty
        self.aligned_others = []

        shape = (self.multispectrum.xsize, self.multispectrum.ysize)
        if self.zero_index is None:
            shf = np.copy(self.index_shift)
        else:
            shf = self.index_shift + self.zero_index

        min_shift = np.nanmin(shf)
        max_shift = np.nanmax(shf)

        # if no shift is observed, then it will not try to apply something
        if min_shift == max_shift:
            print('dataset is already aligned')
            self.aligned = self.multispectrum.copy()
            self.aligned_others = self.other_spectra[:]
            return None

        if self.cropping:
            # set the offset to the maximum shift
            new_offset = self.multispectrum.offset + \
                         self.multispectrum.dispersion * max_shift

            # apply the offset to all the other spectra
            new_offset_list = []
            for spec in self.other_spectra:
                new_offset_list.append(spec.offset +
                                       self.multispectrum.dispersion *
                                       max_shift)

            # the maximum shift should be zero
            shf -= max_shift

            # the new size of the cropped spectrum
            new_size = self.multispectrum.size - (max_shift - min_shift)
            ind = [0, -1 * (max_shift - min_shift)]

            # no odd pixels in spectrum --> For the convolution
            if new_size % 2 == 1:
                new_size -= 1
                ind = [0, -1 * (max_shift - min_shift) - 1]

            ms = MultiSpectrumshape(self.multispectrum.dispersion, new_offset,
                                    new_size, shape[0], shape[1])
            self.aligned = MultiSpectrum(ms, data=np.zeros((shape[0], shape[1],
                                                            new_size)))

            # make new multispectra with the new size of the cropped spectra
            for index, spec in enumerate(self.other_spectra):
                ms = MultiSpectrumshape(self.multispectrum.dispersion,
                                        new_offset_list[index], new_size,
                                        shape[0], shape[1])

                nshape = (shape[0], shape[1], new_size)
                self.aligned_others.append(MultiSpectrum(ms, np.zeros(nshape)))

        else:
            # if no cropping is used, everything is a lot easier :)
            self.aligned = self.multispectrum.copy()

            for spectra in self.other_spectra:
                self.aligned_others.append(spectra.copy())
            ind = [0, None]

        # apply the shifts using the numpy roll function and cropping away
        # the proper regions.
        for index in (np.ndindex(shape)):
            islice = np.s_[index]
            self.aligned.multidata[islice] \
                = np.roll(self.multispectrum.multidata[islice],
                          shf[islice])[ind[0]:ind[1]]

            # apply the shift to every other spectrum
            for ii, spec in enumerate(self.aligned_others):
                spec.multidata[islice] \
                    = np.roll(self.other_spectra[ii].multidata[islice],
                              shf[islice])[ind[0]:ind[1]]

    def align(self):
        """
        Applies the calculated shift to the given multispectrum and other given
        multispectra. The new aligned spectrum is saved in the aligned
        attribute. The additional aligned spectra can be found in the
        aligned_others attribute.
        The difference between the 'align' or 'fast_align' function is that
        align uses the interpolation to shift a spectrum whereas the
        'fast_align' uses the indices to shift so no subpixel shifts are
        possible.

        Returns
        -------
        None.
        """
        if self.shift is None:
            print('The shift value should first be calculated before applying'
                  ' it')
            return None

        # make the aligned others list empty
        self.aligned_others = []

        # determine the minimal and maximum shift
        shape = (self.multispectrum.xsize, self.multispectrum.ysize)

        # some weird prefactor and naming to make everything work
        max_shift = -1*min(np.nanmin(self.shift), 0)
        min_shift = max(np.nanmax(self.shift), 0)

        disp = self.multispectrum.dispersion

        # the zero padded array to make sure that every spectrum can
        # interpolated
        if max_shift == 0:
            ar_before = []
        else:
            ar_before = np.arange(np.ceil(max_shift / disp)) * disp
            ar_before += self.multispectrum.offset-ar_before[-1]-disp

        if min_shift == 0:
            ar_after = []
        else:
            ar_after = np.arange(np.ceil(min_shift / disp))\
                       + self.multispectrum.energy_axis[-1] + disp

        # the new energy axis where everything should be okay.
        E_ax = np.concatenate((ar_before, self.multispectrum.energy_axis,
                               ar_after))

        # cropping modifies the total size EEL spectrum
        if self.cropping:
            ind = [int(np.ceil(max_shift / disp)),
                   -1 * int(np.ceil(min_shift / disp))]

            if (ind[1]-ind[0]) % 2 == 1:
                ind[1] -= 1

            if ind[1] == 0:
                ind[1] = None

            subE = self.multispectrum.energy_axis[ind[0]: ind[1]]
            new_offset = subE[0]
            new_size = subE.size

            delta_offset = self.multispectrum.offset - new_offset

            ms = MultiSpectrumshape(self.multispectrum.dispersion, new_offset,
                                    new_size, shape[0], shape[1])
            self.aligned = MultiSpectrum(ms, data=np.zeros((shape[0], shape[1],
                                                            new_size)))

            # make new multispectra with the new size of the cropped spectra
            for index, spec in enumerate(self.other_spectra):
                prev_off = spec.offset
                ms = MultiSpectrumshape(self.multispectrum.dispersion,
                                        prev_off-delta_offset, new_size,
                                        shape[0], shape[1])
                nshape = (shape[0], shape[1], new_size)
                self.aligned_others.append(MultiSpectrum(ms, np.zeros(nshape)))

        else:
            self.aligned = self.multispectrum.copy()
            for spectra in self.other_spectra:
                self.aligned_others.append(spectra.copy())
            ind = [0, None]

        for index in tqdm(np.ndindex(shape)):
            islice = np.s_[index]
            dat = np.pad(self.multispectrum.multidata[islice],
                         pad_width=(len(ar_before), len(ar_after)))
            si = interpolate.interp1d(E_ax, dat)
            i_dat = si(self.multispectrum.energy_axis+self.shift[islice])
            self.aligned.multidata[islice] = i_dat[ind[0]:ind[1]]

            # apply the shift to every other spectrum
            for ii, spec in enumerate(self.other_spectra):
                dat = np.pad(spec.multidata[islice],
                             pad_width=(len(ar_before), len(ar_after)))
                si = interpolate.interp1d(E_ax, dat)
                i_dat = si(self.multispectrum.energy_axis+self.shift[islice])
                self.aligned_others[ii].multidata[islice] \
                    = i_dat[ind[0]:ind[1]]

    def show_signal_range(self, index=(0, 0)):
        """
        Show the signal range used in the determination of the shift.

        Parameters
        ----------
        index : tuple
            The spectrum of the multispectrum which will be shown.
            (default: (0,0))

        Returns
        -------
        fig: Figure
            The figure which can then be used to modify or save
        """
        ind0 = self.multispectrum.get_energy_index(self.signal_range[0])
        ind1 = self.multispectrum.get_energy_index(self.signal_range[1])
        boolean = np.zeros(self.multispectrum.size)
        boolean[ind0:ind1] = True
        self.multispectrum.setcurrentspectrum(index)
        fig, ax = plt.subplots(figsize=(4, 3))
        ax.plot(self.multispectrum.energy_axis, self.multispectrum.data,
                color='black')
        ax.fill_between(self.multispectrum.energy_axis, 0,
                        self.multispectrum.data.max(), where=boolean,
                        color='green', alpha=0.5)
        ax.set_xlabel(r'Energy loss [eV]')
        return fig

    def show_shift(self, show_index=False, nbins=20):
        """
        Shows the shift which will be applied to the spectra.

        Returns
        -------
        fig: Figure
            The figure which can then be used to modify or save
        """

        if self.multispectrum.xsize == 1 or self.multispectrum.ysize == 1:
            fig, ax = plt.subplots()
            if show_index:
                ax.plot(np.squeeze(self.index_shift))
                ax.set_ylabel(r'Shift [pixel]')
            else:
                ax.plot(np.squeeze(self.shift))
                ax.set_ylabel(r'Shift [eV]')

        else:
            fig, ax = plt.subplots(1, 2, figsize=(6, 3))
            if show_index:
                data = self.index_shift - self.index_shift[0, 0]
                label = 'index'
                bins = np.arange(data.min(), data.max()+1)

            else:
                data = self.shift
                label = 'eV'
                bins = np.linspace(data.min(), data.max(), nbins)

            im = ax[0].imshow(data)
            divider = make_axes_locatable(ax[0])
            cax = divider.append_axes("right", size="5%", pad=0.05)
            cbar = plt.colorbar(im, cax=cax)
            cbar.set_label(label, rotation=270)

            h = np.histogram(data.flatten(), bins=bins)
            ax[1].bar(h[1][:-1], h[0], width=(bins[1]-bins[0]))
            ax[1].set_xlabel('Applied shift [eV]')

        return fig

    def show_alignment_result(self):
        """
        Figure outputs the average spectrum before and after alignment

        Returns
        -------
        fig: Figure
            The figure which can then be used to modify or save
        """
        fig, ax1 = plt.subplots(figsize=(6, 3))
        avg_r = self.multispectrum.mean().data
        avg_a = self.aligned.mean().data
        en_r = self.multispectrum.energy_axis[np.argmax(avg_r)]
        en_a = self.multispectrum.energy_axis[np.argmax(avg_a)]
        intv = 10
        ax1.plot(self.multispectrum.energy_axis, avg_r, label='Raw')
        ax1.plot(self.aligned.energy_axis, avg_a, label='Aligned')
        ax1.legend()
        xlim = [min(en_r, en_a)-intv, max(en_r, en_a)+intv]
        ax1.set_xlim(xlim)

        return fig

    def make_pdf_result(self):
        """
        Creates a pdf which summarizes the results obtained from the
        alignment procedure. This is helpfull to identify a problem
        in this procedure since it is not possible for each multispectrum
        which is acquired.

        Returns
        -------
        fig: Figure
            The figure which can then be used to modify or save
        """
        pdf = PDF(orientation='P', unit='in', format='A4')
        pdf.add_page()

        pdf.intitialize(path=self.pdf_path)
        pdf.write(pdf.dh, txt='Method used is: ' + self.method + '\n')
        pdf.write(0.5, txt=' \n')

        # figure in the resulting shift
        pdf.write(pdf.dh, txt='Cropping was set to: '
                              + str(self.cropping) + '\n')

        dE = np.diff(self.multispectrum.energy_axis[:2])[0]
        pdf.write(pdf.dh, txt='Energy range of raw data is ' + str(dE)
                              + ' eV \n')
        dE = self.aligned.energy_axis[-1] - self.aligned.energy_axis[0]
        pdf.write(pdf.dh, txt='Energy range of the aligned data is ' + str(dE)
                              + ' eV \n')
        pdf.write(0.5, txt=' \n')

        pdf.write(pdf.dh, txt='The shift which is applied to each '
                              'individual spectrum')

        fig = self.show_shift(show_index=False, nbins=20)
        pdf.add_figure(fig)

        # figure on the signal range used
        pdf.add_page()
        pdf.write(pdf.dh, txt='The signal range used in determining the '
                              'shift  \n')
        pdf.write(pdf.dh, txt='Signal range is: '+str(self.signal_range))
        fig1 = self.show_signal_range()
        pdf.add_figure(fig1)

        # figure showing the result
        pdf.add_page()
        pdf.write(pdf.dh, txt='The average raw and aligned spectrum \n')
        fig2 = self.show_alignment_result()
        pdf.add_figure(fig2)

        plt.close('all')
        name = pdf.get_savename()
        pdf.output(name)



import numpy as np
import logging
from pyEELSMODEL.operators.aligns.align import Align

logger = logging.getLogger(__name__)


class FastAlignZeroLoss(Align):
    """
    FastAlignZeroLoss is a class which aligns the zero loss peak and other
    spectra can be added which also need to be aligned. It calculates the
    maximum value of the spectrum.
    """
    def __init__(self, multispectrum, other_spectra=None, signal_range=None,
                 cropping=True):
        """
        Parameters
        ----------
        multispectrum: Multispectrum
            The spectrum from which the background should be removed
        other_spectra: List of Multispectra
            The other multispectra which can be aligned with the same
            parameters
        signal_range: tuple
            Indicates the region on which the zero loss fitting should be
            performed.
        cropping: bool
            Indicates if the edges of the signal are cropped out when aligning
            the zero loss peak.
        """

        super().__init__(multispectrum, other_spectra, cropping, signal_range,
                         zero_index=0)
        self.method = 'ZLP Index Maximum'

    def determine_fast_shift(self):
        """
        Sets the index_shift and shift attribute by finding the maximum value
        in each spectrum.

        Returns
        -------
        None.

        """
        ind0 = self.multispectrum.get_energy_index(self.signal_range[0])
        ind1 = self.multispectrum.get_energy_index(self.signal_range[1])
        co = np.argmax(self.multispectrum.multidata[:, :, ind0:ind1], axis=2)
        shift = -1 * (co + ind0)
        self.shift = self.multispectrum.energy_axis[-1 * shift]
        self.index_shift = shift

    def perform_alignment(self):
        """
        Performs the alignment procedure by first calculating the shift and
        then applying it to the spectra. The roll method is chosen to perform
        the alignment.

        Returns
        -------
        None.

        """
        self.determine_fast_shift()
        self.fast_align()
        # self.align()



import numpy as np
import logging
from tqdm import tqdm
from pyEELSMODEL.core.operator import Operator
from pyEELSMODEL.core.spectrum import Spectrum, Spectrumshape
from pyEELSMODEL.core.model import Model

from pyEELSMODEL.components.powerlaw import PowerLaw
from pyEELSMODEL.components.CLedge.zezhong_coreloss_edgecombined import \
    ZezhongCoreLossEdgeCombined
from pyEELSMODEL.components.MScatter.mscatterfft import MscatterFFT
from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.components.plasmon import Plasmon
from pyEELSMODEL.components.fixedpattern import FixedPattern
from pyEELSMODEL.core.multispectrum import MultiSpectrumshape, MultiSpectrum

logger = logging.getLogger(__name__)


class CoreLossSimulator(Operator):
    """
    Coreloss EELS spectrum simulator for maps. This uses the atomic cross
    sections for the edge shapes. The simulated low loss is consists out of the
    zer-loss peak and the bulk plasmons. Note that the low loss is only used
    to modify the shape of the edges and the background. Hence the real shape
    is not so important. The background in the core-loss is found
    by convolving a powerlaw with the background to include multiple
    scattering.

    """
    def __init__(self, multispecshape, elements, edges, maps, tlambda_map,
                 settings):
        """

        Parameters
        ----------
        multispecshape: MultiSpectrumShape
            The multispectrum form which an area need to be selected
        elements: list of strings
            The elements list should contain the elements as indicated on the
            periodic table
        edges: list of strings
            The edges list should contain the letters of the edges
            (K, L, M, ...)
        settings: tuple
            Tuple containing E0 (acceleration voltage), alpha (convergence
             angle) and beta (collection angle).
        ll: Spectrum or MultiSpectrum
            The low loss which can be added. This will be used to align the
            spectra with each other and will be used to model the multiple
            scattering.
        """
        self.spectrumshape = multispecshape.spectrumshape
        self.scan_size = (multispecshape.xsize, multispecshape.ysize)
        self.size = multispecshape.Esize
        self.ll = None

        self.elements = elements
        self.edges = edges
        self.onsets = np.zeros((len(elements)))
        self.maps = maps  # contains number of elements 2d maps

        self.E0 = settings[0]
        self.alpha = settings[1]
        self.beta = settings[2]
        self.qsteps = 20  # the number of points used to integrate q vector

        # low loss settings
        self.tlambdas = tlambda_map
        # energy of bulk plasmon at each probe position
        self.Ep_map = 22*np.ones_like(self.tlambdas)
        # width of the bulk plasmon at each probe position
        self.Ew_map = 5*np.ones_like(self.tlambdas)
        self.n_plasmon = 5  # number of plasmon included
        self.fwhm = 1  # eV
        self.llspectrumshape = Spectrumshape(multispecshape.dispersion,
                                             -self.fwhm*10,
                                             multispecshape.Esize)

        # background settings
        self.A_bg = 1e3  # amplitude background
        self.r_bg = 3  # powerlow
        self.bg_method = 'convolved'  # indicates if background needs to be
        # convolved with the low loss. This gives a more accurate description
        # of reality.
        # to validate the CRLB, it is usefull to know the exact shape of the
        # background so this convolution can be turned off.
        self.noise = 1  # noise level

        # shift settings
        self.use_shift = False

        # detector + noise
        self.add_poisson = True  # indicates if poisson noise needs to be added

    def make_lowloss(self):
        """
        Creates the low loss multispectrum using the information on the zero-
        loss peak (fwhm) and bulk plasmon (Ep, tlambda, Ew, beta).

        """
        zlp = Gaussian(self.llspectrumshape, A=1, centre=0, fwhm=self.fwhm)
        zlp.calculate()

        shape = self.scan_size + (self.spectrumshape.size,)
        lldata = np.zeros(shape)
        for index in tqdm(np.ndindex(self.scan_size)):
            islice = np.s_[index]
            A = zlp.data.sum()
            plasmon = Plasmon(self.llspectrumshape, A=A,
                              Ep=self.Ep_map[islice], Ew=self.Ew_map[islice],
                              n=self.n_plasmon,
                              tlambda=self.tlambdas[islice], beta=self.beta,
                              E0=self.E0)
            plasmon.calculate()
            lldata[islice] = zlp.data + plasmon.data
            # lldata[islice] = zlp.data
        llsh = MultiSpectrumshape(self.llspectrumshape.dispersion,
                                  self.llspectrumshape.offset,
                                  self.llspectrumshape.size,
                                  self.scan_size[0],
                                  self.scan_size[1])
        ll = MultiSpectrum(llsh, data=lldata)

        self.ll = ll

    def _calculate_convoluted_background(self, llcomp):
        """
        Calculates a background which is convoluted with the low loss.
        This represents the difference in background when thickness
        changes for the same core loss spectrum.

        Returns
        -------
        bg : FixedPattern
            Background from which only the amplitude can be modified.
            This background is also set to not convolute when applied
            into a model.

        """
        ndispersion = self.spectrumshape.dispersion
        noffset = max(self.spectrumshape.offset -
                      0.5 * self.spectrumshape.size
                      * self.spectrumshape.dispersion,
                      self.spectrumshape.dispersion)

        # nspecshape = Spectrumshape(ndispersion,noffset,2*specshape.size)
        nspecshape = Spectrumshape(ndispersion, 100,
                                   2 * self.spectrumshape.size)

        bkg = PowerLaw(nspecshape, self.A_bg, self.r_bg)
        bkg.canconvolute = True
        bkg.calculate()

        energy_axis = np.arange(nspecshape.size) * ndispersion + noffset
        bkgs = Spectrum.from_numpy(bkg.data, energy_axis)
        mspecshape = bkgs.get_spectrumshape()

        llspec = Spectrum.from_numpy(llcomp.data, llcomp.energy_axis)
        llspec.dispersion = mspecshape.dispersion
        llcompg = MscatterFFT(mspecshape, llspec)

        mod = Model(mspecshape, [bkg, llcompg])

        mod.calculate()

        bg = FixedPattern(self.spectrumshape, mod)
        # print(bg.parameters[0].getvalue())
        bg.calculate()
        # bg.plot()
        bg.parameters[0].setvalue(self.A_bg / bg.data[0])
        bg.calculate()
        bg.canconvolute = False
        return bg

    def get_background(self, ll):
        """
        Return the background dependent on the bg_method attribute.

        Parameters
        ----------
        ll: MultiSpectrum
            The low loss multispectrum.

        """
        if self.bg_method == 'convolved':
            bg = self._calculate_convoluted_background(ll)
        else:
            bg = PowerLaw(self.spectrumshape, self.A_bg, self.r_bg)
        return bg

    def make_elements(self):
        """
        Makes a list containing the atomic core-loss edges.

        """
        comp_elements = []
        for elem, edge, onset in zip(self.elements, self.edges, self.onsets):
            comp = ZezhongCoreLossEdgeCombined(self.spectrumshape, 1, self.E0,
                                               self.alpha,
                                               self.beta, elem, edge,
                                               eshift=onset,
                                               q_steps=self.qsteps)

            comp_elements.append(comp)
        self.element_components = comp_elements

    def make_coreloss(self):
        """
        Calculated the core-loss multispectrum.

        """
        shape = self.scan_size + (self.spectrumshape.size,)
        cldata = np.zeros(shape)

        for index in tqdm(np.ndindex(self.scan_size)):
            islice = np.s_[index]
            llcomp = MscatterFFT(self.spectrumshape, self.ll[islice])
            bg = self.get_background(self.ll[islice])
            for jj, comps in enumerate(self.element_components):
                comps.parameters[0].setvalue(self.maps[jj][islice])

            self.mod = Model(self.spectrumshape,
                             components=[bg]+self.element_components+[llcomp])
            self.mod.calculate()
            if self.add_poisson:
                cldata[islice] = np.random.poisson(self.noise * self.mod.data)
            else:
                cldata[islice] = self.noise * self.mod.data

        s = MultiSpectrum.from_numpy(cldata, self.mod.energy_axis)
        self.multispectrum = s

    def add_shift(self):
        """
        Shifts both low loss and core loss with a sine function.
        The amplitude is 0.5% of the energy range and the period is chosen such
        that 20 periods are in the STEM-EELS map.
        """
        Ashift = 0.005*(self.ll.energy_axis[-1] - self.ll.energy_axis[0])
        period = (self.ll.xsize * self.ll.ysize)/20
        x = np.arange(self.ll.xsize*self.ll.ysize)
        shift1d = Ashift*np.sin(2*np.pi*x/period)
        self.shift2d = np.reshape(shift1d,
                                  (self.ll.xsize, self.ll.ysize)).astype('int')

        cldata = np.zeros_like(self.multispectrum.multidata)
        lldata = np.zeros_like(self.ll.multidata)

        for index in tqdm(np.ndindex(self.scan_size)):
            islice = np.s_[index]
            cldata[islice] = np.roll(self.multispectrum.multidata[islice],
                                     self.shift2d[islice])
            lldata[islice] = np.roll(self.ll.multidata[islice],
                                     self.shift2d[islice])

        cldata = cldata[:, :, self.shift2d.max():-self.shift2d.max()]
        lldata = lldata[:, :, self.shift2d.max():-self.shift2d.max()]

        cl_offset = self.multispectrum.offset +\
            self.shift2d.min()*self.multispectrum.dispersion
        cl_energy = np.arange(cldata.shape[2]) * \
            self.multispectrum.dispersion + cl_offset

        ll_offset = self.ll.offset + self.shift2d.min() * self.ll.dispersion
        ll_energy = np.arange(cldata.shape[2]) * self.ll.dispersion + ll_offset

        self.multispectrum = MultiSpectrum.from_numpy(cldata, cl_energy)
        self.ll = MultiSpectrum.from_numpy(lldata, ll_energy)

    def simulate_multispectrum(self):
        self.make_lowloss()
        self.make_elements()
        self.make_coreloss()

        if self.use_shift:
            self.add_shift()

        print('Multispectrum is simulated')



from pyEELSMODEL.components.MScatter.mscatter import Mscatter
import numpy as np


class MscatterFFT(Mscatter):
    """
    Mutiple scattering using FFT (e.g. to concolve model with LL spectrum)
    """

    def __init__(self, specshape, llspectrum, use_padding=True,padding_mode_data="constant",padding_mode_llspectrum="constant"):
        """
        Parameters
        ----------
        specshape: Spectrumshape
            The spectrum shape on the spectrum it will be used, not the low
            loss spectrum
        llspectrum: Spectrum or MultiSpectrum
            The spectrum or multispectrum which is used to convolve the rest
            of the components with.
        use_padding: bool
            Indicates if both spectrum data and low loss are zero padded to
            reducde artifacts coming from the FFT. If True, the calculations
            take longer but are more precise. (default: True)

        padding_mode_data: string
        Padding mode to use on data, see the different modes in np.pad. (Recommended: "edge")

        padding_mode_llspectrum: string
        Padding mode to use on llspectrum, see the different modes in np.pad. (Recommended: "constant")


        """
        super().__init__(specshape, llspectrum)
        self._setname("Multiple scattering (FFT)")
        self.setdescription("Convolution of the HL spectrum with LL using fast"
                            " fourier transform convolution.\nThis simulates"
                            " the effect of multiple scattering\nwhich is an "
                            "important effect in experimental spectra ")
        self.padding_mode_data=padding_mode_data
        self.padding_mode_llspectrum = padding_mode_llspectrum
        self.padding = llspectrum.size
        self.use_padding = use_padding
        # some cache to not recalculate fourier transfrom when not needed
        # for instance in the calculation of convolved A matrix
        self.new_ll = True

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        if self.use_padding:
            self.calculate_w_padding()
        else:
            self.calculate_raw()

    def calculate_raw(self):
        fmodel = np.fft.rfft(self.data)  # real fourier transform the model
        if self.new_ll:
            self.llspectrum.normalise()
            fll = np.fft.rfft(
                self.llspectrum.data)  # real fourier transform the ll spectrum
            # shift zl peak position to 0!
            # need to compensate for zl peak not being at pix 0
            self.zlindex = self.llspectrum.getmaxindex()
            self.fll = fll
        self.data = np.roll(np.fft.irfft(fmodel * self.fll), -self.zlindex)

    def calculate_w_padding(self):
        """
        Function which adds the zero padding to remove the intensity of the
        end of the model to come into the beginning of the model.
        """
        pds = (self.padding, self.padding)
        # real fourier transform the model
        fmodel = np.fft.rfft(np.pad(self.data, pad_width=pds,mode = self.padding_mode_data))
        if self.new_ll:
            self.llspectrum.normalise()
            llpad = np.pad(self.llspectrum.data, pad_width=pds,mode = self.padding_mode_llspectrum)
            fll = np.fft.rfft(llpad)  # real fourier transform the ll spectrum
            self.fll = fll
            self.zlindex = np.argmax(llpad)

        # conv = np.real(np.fft.ifft(fmodel*fll)[self.padding:-self.padding])
        # conv = np.real(np.fft.ifft(fmodel * self.fll))
        conv = np.fft.irfft(fmodel * self.fll)

        # shift zl peak position to 0!
        # need to compensate for zl peak not being at pix 0
        self.data = np.roll(conv, -self.zlindex)[self.padding:-self.padding]

    def calculate_A_matrix(self, A_matrix):
        """
        Convolution via matrices instead of for loop.
        This function is not faster than the regular convolution and is not
        used in pyEELSMODEL.
        """
        dpad = ((self.padding, self.padding), (0, 0))
        A_matrix_pad = np.pad(A_matrix, pad_width=dpad)
        fftA = np.fft.fft(A_matrix_pad, axis=0)

        self.llspectrum.normalise()
        llpad = np.pad(self.llspectrum.data,
                       pad_width=(self.padding, self.padding))
        self.zlindex = np.argmax(llpad)

        fll = np.fft.fft(llpad)  # real fourier transform the ll spectrum
        fll_matrix = np.tile(fll[:, np.newaxis], (1, A_matrix.shape[1]))

        conv = np.real(np.fft.ifft(fftA * fll_matrix, axis=0))
        conv_A = np.roll(conv, -self.zlindex,
                         axis=0)[self.padding:-self.padding, :]
        return conv_A



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 29 16:25:55 2021

@author: joverbee
"""

import sys
from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.core.spectrum import Spectrumshape, Spectrum
from pyEELSMODEL.core.fitter import Fitter
from pyEELSMODEL.core.model import Model
import numpy as np
sys.path.append("..")  # Adds higher directory to python modules path.


def test_gaussian():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = Gaussian(specshape, 100, 500, 10)  # create a component
    bkg.calculate()

    sigma = np.abs(10) / (2.0 * np.sqrt(2.0 * np.log(2.0)))
    gauss = 100*np.exp(-0.5*((bkg.energy_axis-500)/sigma)**2)
    assert np.all(np.abs(gauss-bkg.data) < 1e-6)


def test_gaussian_gradient():
    specshape = Spectrumshape(1, 100, 1024)  # dispersion, offset, size
    bkg = Gaussian(specshape, 100, 500, 10)  # create a component
    bkg.calculate()

    gradA = bkg.getgradient(bkg.parameters[0])
    sigma = np.abs(10) / (2.0 * np.sqrt(2.0 * np.log(2.0)))
    gauss = 1*np.exp(-0.5*((bkg.energy_axis-500)/sigma)**2)
    assert np.all(np.abs(gradA-gauss)) < 1e-6

    # test the gradients of the gaussian by comparing to the numerical
    # derivatives
    mod = Model(specshape)
    mod.addcomponent(bkg)
    s = Spectrum(specshape, data=np.ones(specshape.size))
    fit = Fitter(s, mod)

    dat = fit.numerical_partial_derivative(bkg.parameters[1])
    grad = bkg.getgradient(bkg.parameters[1])
    assert np.all(np.abs(dat-grad) < 1.5)

    dat = fit.numerical_partial_derivative(bkg.parameters[2])
    grad = bkg.getgradient(bkg.parameters[2])

    # plt.figure()
    # plt.plot(dat-grad)
    # plt.plot(dat)
    assert np.all(np.abs(dat - grad) < 0.01)


def main():
    test_gaussian()
    test_gaussian_gradient()


if __name__ == "__main__":
    main()



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy import ndimage
import h5py
from os.path import exists
import os
from tqdm import tqdm

from pyEELSMODEL.io_tools.dm_ncempy import dmReader
from pyEELSMODEL.io_tools.hdf5_io import load_h5py, load_hspy
from pyEELSMODEL.core.spectrum import Spectrum, Spectrumshape

import copy
import logging

logger = logging.getLogger(__name__)


class MultiSpectrumshape:
    """
    MultiSpectrumshape is a class holding the main parameters of a spectrum in
    order to compactly create several spectra with the same size by providing
    an instance of this class.
    It holds:
        dispersion: float, dispersion in eV/pixel describing the energy scale
        offset: float, energy of the first pixel in the spectrum in eV
        size: int, the size of a spectrum in number of pixels (eg typ 1024)
        xsize: int number of spectra in x direction
        ysize: int number of spectra in y direction
    """

    def __init__(self, dispersion, offset, Esize, xsize, ysize):
        self.dispersion = dispersion
        self.offset = offset
        self.Esize = Esize
        self.xsize = xsize
        self.ysize = ysize
        self.spectrumshape = Spectrumshape(dispersion, offset, Esize)
        self._index = 0
        logger.debug('Multispectrumshape init')

    def getspectrumshape(self):
        return self.spectrumshape


class MultiSpectrumIterator:
    """MultiSpectrum Iterator class"""

    def __init__(self, multispectrum):
        # Team object reference
        self._multispectrum = multispectrum
        # member variable to keep track of current index
        self._index = 0
        self.size = multispectrum.xsize * multispectrum.ysize

    def __next__(self):
        if self._index < self.size:
            result = self._multispectrum[self._index]
            self._index += 1
            return result
        raise StopIteration


class MultiSpectrum(Spectrum):
    def __init__(self, multispectrumshape, data=None, acq_time=0):
        """
        Initialises a MultiSpectrum instance

        Parameters
        ----------
        spectrumshape : MultiSpectrumshape
            holds the basic parameters dispersion [eV/pixel], offset [eV] and
            size [pixels] to create the energy scale of the spectrum
        data : float or int, optional
            Holds the EELS intensity data with same size as spectrumshape.size.
            If None is given an empty spectrum is created with all zeros. The
            default is None.

        Raises
        ------
        ValueError
            if data does not have same size as in spectrumshape.size
        TypeError
            if data is neither int or float

        Returns
        -------
        An instance of a Spectrum.

        """
        spectrumshape = Spectrumshape(multispectrumshape.dispersion,
                                      multispectrumshape.offset,
                                      multispectrumshape.Esize)

        super().__init__(spectrumshape, acq_time=acq_time)

        self.xsize = multispectrumshape.xsize
        self.ysize = multispectrumshape.ysize
        self.currentspectrumid = (0, 0)

        if data is None:
            # create empy vector of data
            self.multidata = np.zeros((self.xsize, self.ysize, self.size))

        else:
            sh = (self.xsize, self.ysize, multispectrumshape.Esize)
            if data.shape != sh:
                raise ValueError('data needs to be same size as spectrum.')
            if not np.issubdtype(data.dtype, np.floating):
                raise TypeError('data needs to be convertible to float.')
            self.multidata = data

        self.setcurrentspectrum(self.currentspectrumid)
        self.setcurrentmeanspectrum(self.currentspectrumid, 1, 1)

        # Array which points to exclude
        self.exclude = np.zeros(self.size, dtype=bool)
        self.pppc = 1.0
        self.name = 'a multispectrum'

    @classmethod
    def load(cls, filename, flip_sign=False):
        ext = os.path.splitext(filename)[-1]
        if ext == '.hdf5':
            s = cls.load_hdf5(filename)

        elif ext == '.hspy':
            s = cls.load_hspy(filename)

        elif (ext == '.dm3') or (ext == '.dm4'):
            s = cls.load_dm(filename, flip_sign=flip_sign)

        else:
            raise ValueError(r'Extension is not valid')

        return s

    @classmethod
    def load_hdf5(cls, filename):
        """
        Loads the multispectrum from .hdf5 file

        Parameters
        ----------
        filename : string
              Filename containing the data.

        Returns
        ----------
        s: MultiSpectrum
            The multispectrum which is inside the filename

        """
        data, dispersion, offset, size = load_h5py(filename)
        specshape = MultiSpectrumshape(dispersion, offset, size, data.shape[0],
                                       data.shape[1])
        s = MultiSpectrum(specshape, data=data)

        # how to return alternate extra data like haadf map?
        return s

    @classmethod
    def load_hspy(cls, filename):
        """
        Loads hyperspy data file format.

        Parameters
        ----------
        filename : string
              Filename containing the data.

        Returns
        ----------
        s: MultiSpectrum
            The multispectrum which is inside the filename
        df: list
            List containing other acquired data

        """
        params, df, detector_type = load_hspy(filename)
        specs = []
        for param in params:
            specshape = MultiSpectrumshape(param[1], param[2], param[3],
                                           param[0].shape[0],
                                           param[0].shape[1])
            s = MultiSpectrum(specshape, data=param[0], acq_time=param[4])
            specs.append(s)
        return specs, df

    @classmethod
    def load_dm(cls, filename, flip_sign=False, dispersion=None):
        """
        Loading dm data

        Parameters
        ----------
        filename : string
              Filename containing the data.
        flip_sign: boolean
            Bug when loading in data where the energy axis starts negative.
            If wrong is set to True, then this bug is worked around.
        dispersion: float
            Sometimes the dispersion in dual EELS is different, this is a
            method to force it to be the same.

        Returns
        ----------
        s: MultiSpectrum
            The multispectrum which is inside the filename
        """
        dmfile = dmReader(filename)
        print(dmfile['data'].ndim)
        if dmfile['data'].ndim == 3:
            data = np.swapaxes(dmfile['data'], 0, 2)
            e_axis = dmfile['coords'][0]

        elif dmfile['data'].ndim == 2:
            data = dmfile['data'][:, np.newaxis, :]
            e_axis = dmfile['coords'][1]

        multi_specshape = MultiSpectrumshape(e_axis[1] - e_axis[0], e_axis[0],
                                             e_axis.size, data.shape[0],
                                             data.shape[1])
        ms = MultiSpectrum(multi_specshape, data=data)
        if dispersion is not None:
            ms.dispersion = dispersion

        # bug which is something that you can manually solve.
        if flip_sign:
            ms.offset = -1 * ms.offset

        return ms

    @classmethod
    def from_numpy(cls, data_array, energy_axis):
        """
        Creates an Multispectrum object from the data and energy axis using
        numpy arrays.
        Usefull when playing with different functions which are not (yet)
        integrated into pyEELSmodel functionalities.

        Parameters
        ----------
        data_array : numpy array (1D)
            The data of the EEL spectrum.
        energy_axis: numpy array (1D)
            The energy axis used [eV]

        Returns
        ----------
        s: MultiSpectrum
            The multispectrum which is inside the filename

        """
        dispersion = energy_axis[1] - energy_axis[0]
        offset = energy_axis[0]
        size = energy_axis.size
        xsize = data_array.shape[0]
        ysize = data_array.shape[1]
        specshape = MultiSpectrumshape(dispersion, offset, size, xsize, ysize)
        s = MultiSpectrum(specshape, data=data_array)
        return s

    @property
    def multidata(self):
        return self._multidata

    @multidata.setter
    def multidata(self, multidata):
        self._multidata = multidata
        self.setcurrentspectrum((0, 0))

    def indexOK(self, id):
        """
        Index is valid for the scan dimensions of the EELS map

        Parameters
        ----------
        id : tuple int
              Check if the given index is valid.

        Returns
        ----------
        bool:
            Returns True if the index is valid.

        """
        if (0 <= id[0] < self.xsize) and (0 <= id[1] < self.ysize):
            return True
        return False

    def get_multispectrumshape(self):
        sh = MultiSpectrumshape(self.dispersion, self.offset, self.size,
                                self.xsize, self.ysize)
        return sh

    def __getitem__(self, key):
        """
        [] operator, makes a copy of the multispectrum where key
        is the indexing used.

        Parameters
        ----------
        key : indexing
            Indexing of the multispectrum can be performed as is done with
            numpy arrays.

        Raises
        ------
        ValueError
            DESCRIPTION.

        Returns
        -------
        s: MultiSpectrum/Spectrum
            The sliced multispectrum or spectrum
        TODO add some warining when slicing the energy loss spectrum because
        this is something we do not want
        """

        ndata = self.multidata[key]

        # check if there is an int in the key list and make sure to add this
        # dimension in the data
        for i, val in enumerate(key):
            if isinstance(val, int):
                ndata = np.expand_dims(ndata, axis=i)

        # check if the energy axis is sliced
        if ndata.shape[-1] != self.size:
            if key[-1].start is None:
                offset = self.offset
            else:
                offset = self.energy_axis[key[-1].start]
            if key[-1].step is None:
                dispersion = self.dispersion
            else:
                dispersion = self.dispersion * key[-1].step
            size = ndata.shape[-1]
        else:
            offset = self.offset
            dispersion = self.dispersion
            size = self.size

        # return a single spectrum and not multispectrum when only one spectrum
        # is selected
        if (ndata.shape[0] == 1) and (ndata.shape[1] == 1):
            ms = Spectrumshape(dispersion, offset, size)
            return Spectrum(ms, data=ndata[0, 0])
        else:
            ms = MultiSpectrumshape(dispersion, offset, size, ndata.shape[0],
                                    ndata.shape[1])
            return MultiSpectrum(ms, data=ndata)

    def sum(self, axis=(0, 1)):
        """
        Sum the real space coordinates with each other. When summed over
        both direction, a single spectrum is returned instead of a
        multispectrum

        Parameters
        ----------
        axis : tuple
            The axis over which the sum needs to be performed.

        Returns
        -------
        m: MultiSpectrum/Spectrum
            The summed multispectrum or spectrum
        """
        # Cannot sum over the energy axis only the real space coordinates
        if isinstance(axis, int):
            if axis == 2:
                raise ValueError(r'Cannot sum over the energy axis')
        else:
            for el in axis:
                if el == 2:
                    raise ValueError(r'Cannot sum over the energy axis')

        ndata = self.multidata.sum(axis=axis)

        if ndata.ndim == 1:
            ms = Spectrumshape(self.dispersion, self.offset, self.size)
            m = Spectrum(ms, data=ndata)

        else:
            ndata = np.expand_dims(ndata, axis)
            ms = MultiSpectrumshape(self.dispersion, self.offset, self.size,
                                    ndata.shape[0], ndata.shape[1])
            m = MultiSpectrum(multispectrumshape=ms, data=ndata)

        m.exclude = self.exclude[:]  # todo nicer way of copying the exclude
        return m

    def mean(self, axis=(0, 1)):
        """
        Mean the real space coordinates with each other. When summed over
        both direction, a single spectrum is returned instead of a
        multispectrum

        Parameters
        ----------
        axis : tuple
            The axis over which the mean needs to be performed.

        Returns
        -------
        s: MultiSpectrum/Spectrum
            The
            summed multispectrum or spectrum
        """
        if axis == (0, 1):
            n_spec = self.xsize * self.ysize
        elif axis == 0:
            n_spec = self.xsize
        elif axis == 1:
            n_spec = self.ysize

        s = self.sum(axis)
        s.data = s.data / n_spec
        return s

    def integrate(self, window=None, index_type=False):
        """
        Integrates the signal over the given energy window or
        the given index.

        Parameters
        ----------
        window : tuple
            The window over which to integrate. If index_type is True,
            the integration window is given with the index instead of energy.
            If window is None, the integral over the entire energy range is
            given.
        index_type: bool
            Indicates if the window should be interpreted as the indices or
            the energy of the spectrum

        Returns
        -------
        result : numpy array of float of sizex,sizey
            The value of the integration

        """
        if window is None:
            window = [0, self.size]
            index_type = True

        if index_type:
            result = self.multidata[:, :, window[0]:window[1]].sum(-1)
        else:
            ind0 = self.get_energy_index(window[0])
            ind1 = self.get_energy_index(window[1])
            result = self.multidata[:, :, ind0:ind1].sum(-1)
        return result

    def setcurrentspectrum(self, index):
        """
        Sets the current spectrum to the given index

        Parameters
        ----------
        index : tuple
            The index to which the multispectrum is set.

        Returns
        -------
        None

        """
        if self.indexOK(index):
            self.currentspectrumid = index
            self.data = self.multidata[index[0], index[1], :]
        else:
            raise IndexError('Index out of bounds multispectrum')

    def setcurrentmeanspectrum(self, index, width, height):
        """
        Sets the current mean spectrum to the given index using the proper
        width and height. Usefull for the GUI.

        Parameters
        ----------
        index : tuple
            The index to which the multispectrum is set.

        Returns
        -------
        None

        """
        if self.indexOK(index):
            xx = [index[0], index[0] + width]
            yy = [index[1], index[1] + height]
            ndata = self.multidata[xx[0]:xx[1], yy[0]: yy[1], :].mean((0, 1))
            self.meandata = ndata
        else:
            raise IndexError('Index out of bounds multispectrum')

    def __iter__(self):
        return MultiSpectrumIterator(self)

    def swap_xy(self):
        """
        Function which will swap the x and y coordinates of the scan
        and return this as a new multispectrum.

        :return:
        """
        sh = MultiSpectrumshape(self.dispersion, self.offset, self.size,
                                self.ysize, self.xsize)
        s = MultiSpectrum(sh, data=np.swapaxes(self.multidata, 0, 1))
        return s

    def copy(self):
        """
        Returns a copy of the object

        """
        return copy.deepcopy(self)

    def geteshift(self):
        """
        Get energy shift. This shift allows to move the whole energy scale up
        and down by calling seteshift(energy)

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.eshift

    def __mul__(self, other):
        """
        Multiply two spectra with each other. All properties of the self
        spectrum are copied, except the data is multiplied with the data of
        the other spectrum. If other is an int or float, the spectrum is
        multiplied by that number.

        Parameters
        ----------
        other : MultiSpectrum
            The multispectrum to multiply with.

        Raises
        ------
        ValueError
            In case both spectra don't have the same spectrumshape settings
            (in which case multiply makes no sense).
        TypeError
            In case other is neither a Spectrum, int or float.

        Returns
        -------
        s : Spectrum
            Returns a reference to a -new- spectrum that is the multiplication
            result.

        """
        if isinstance(other, MultiSpectrum):
            if self._check_not_same_settings(other):
                raise ValueError('Dispersion, offset or size are not the same')

            else:
                s = self.copy()
                s.data = self.data * other.data
                s.multidata = self.multidata * other.multidata
                return s

        else:
            raise TypeError('Input should be spectrum object, float or int')

    def __add__(self, other):
        """
        Add two spectra. All properties of the self spectrum are
        copied, except the data is added with the data of the other spectrum.
        If other is an int or float, the number is added to that spectrum.

        Parameters
        ----------
        other : MultiSpectrum
            The multispectrum to add.

        Raises
        ------
        ValueError
            In case both spectra don't have the same spectrumshape settings
            (in which case adding makes no sense).
        TypeError
            In case other is neither a Spectrum, int or float.

        Returns
        -------
        s : MultiSpectrum
            Returns a reference to a -new- spectrum that is the addition
             result.

        """
        if isinstance(other, MultiSpectrum):
            if self._check_not_same_settings(other):
                raise ValueError('Disperion, offet or size are not the same')

            else:
                s = self.copy()
                s.data = self.data + other.data
                s.multidata = self.multidata + other.multidata
                return s

        else:
            raise TypeError('Input should be spectrum object, float or int')

    def __sub__(self, other):
        """
        Subtract two spectra. All properties of the self spectrum are
        copied, except the data is added with the data of the other spectrum.
        If other is an int or float, the number is subtracted from that
        spectrum.

        Parameters
        ----------
        other : MultiSpectrum
            The multispectrum to subtract.

        Raises
        ------
        ValueError
            In case both spectra don't have the same spectrumshape settings
            (in which case subtracting makes no sense).
        TypeError
            In case other is neither a Spectrum, int or float.

        Returns
        -------
        s : Spectrum
            Returns a reference to a -new- spectrum that is the subtraction
            result.

        """
        if isinstance(other, MultiSpectrum):
            if self._check_not_same_settings(other):
                raise ValueError('Dispersion, offset or size are not the same')

            else:
                s = self.copy()
                s.data = self.data - other.data
                s.multidata = self.multidata - other.multidata

                return s

        else:
            raise TypeError('Input should be MultiSpectrum object, float or '
                            'int')

    def __truediv__(self, other):
        return self.__div__(other)

    def __div__(self, other):
        """
        Divide two spectra by each other. All properties of the self spectrum
        are copied, except the data is divided with the data of the other
        spectrum. If other is an int or float, the spectrum is divided by that
        number.

        Parameters
        ----------
        other : MultiSpectrum
            The spectrum to divide by.

        Raises
        ------
        ValueError
            In case both spectra don't have the same spectrumshape settings
            (in which case divide makes no sense).
        TypeError
            In case other is neither a Spectrum, int or float.

        Returns
        -------
        s : Spectrum
            Returns a reference to a -new- spectrum that is the division
            result.

        """
        if isinstance(other, MultiSpectrum):
            if self._check_not_same_settings(other):
                raise ValueError('Dispersion, offset or size are not the same')

            else:
                s = self.copy()
                s.data = self.data / other.data
                s.multidata = self.multidata / other.multidata

                return s

        else:
            raise TypeError('Input should be spectrum object, float or int')

    def get_interval(self, interval, even_size=True):
        """
        Parameters
        ----------
        interval : tuple
            Interval which has the start and end of the sub spectrum
        even_size: bool
            Makes the size of the returned multispectrum even to accomodate for
             a fast low loss convolution
        Raises
        ------
        ValueError
            In case both spectra don't have the same spectrumshape settings
            (in which case divide makes no sense).
        TypeError
            In case other is neither a Spectrum, int or float.

        Returns
        -------
        s : MultiSpectrum
            Returns a reference to a -new- spectrum that is the division
            result.

        """

        ind0 = self.get_energy_index(interval[0])
        ind1 = self.get_energy_index(interval[1])

        is_even = (ind1 - ind0) % 2
        if (is_even != 0) & even_size:
            print('Make the size of the energy spectrum even')
            ind1 -= 1

        ndata = self.multidata[:, :, ind0:ind1]
        noffset = self.offset + ind0 * self.dispersion
        sh = MultiSpectrumshape(self.dispersion, noffset, ndata.shape[2],
                                self.xsize, self.ysize)
        s = MultiSpectrum(sh, data=ndata)
        return s

    def _check_not_same_settings(self, other):
        """
        Checks if the other spectrum has the same spectrumshape properties as
        self.

        Parameters
        ----------
        other : Spectrum
            A spectrum to compare with.

        Returns
        -------
        bool
            True if both spectra have the same spectrumshape, False otherwise.

        """

        return (self.dispersion != other.dispersion) or (
                    self.offset != other.offset) or (
                           self.size != other.size) or (
                           self.xsize != other.xsize) or (
                           self.ysize != other.ysize)

    #   def _print_type_warning(self):
    #       print('Input should be spectrum object, float or int')

    def plot(self, index=None, externalplt=None, **kwargs):
        """
        Plot the spectrum

        Parameters
        ----------
        externalplt : matplotlib reference
              A reference to an external matplotlib reference, if None we use
              our own matplotlib and create a new figure.
        """
        tempplt = plt
        if isinstance(externalplt, plt.Figure):
            tempplt = externalplt
        else:
            # create our own figure
            plt.figure()

        if index is not None:
            self.setcurrentspectrum(index)

        tempplt.plot(self.energy_axis, self.data, **kwargs)
        tempplt.xlabel(r'Energy Loss [eV]')
        tempplt.ylabel('Counts')
        tempplt.title(self.currentspectrumid)

    def show_excluded_region(self, **kwargs):
        fig, ax = plt.subplots()
        ax.plot(self.energy_axis, self.mean().data, color='black')
        ax.fill_between(self.energy_axis, 0, self.mean().data.max(),
                        where=self.exclude, color='green', alpha=0.5)

    def erase(self):
        self.data = np.zeros(self.size)

    def gaussian_3d(self, X, Y, Z, sigma):
        """
        Computes the 3 dimensional gaussian where sigma can be non isotropic.

        Parameters
        ----------
        X : numpy array (3D)
            Meshgrid indicating the x coordinate of the mesh.
        Y : numpy array (3D)
            Meshgrid indicating the x coordinate of the mesh.
        Z : numpy array (3D)
            Meshgrid indicating the x coordinate of the mesh.
        sigma : tuple of floats
            Tuple of length 3 of the sigmas of each gaussian.
        Returns
        -------
        f: numpy array (3D)
            Three dimensional gaussian.

        """
        f = np.exp(
            -((X ** 2 / (2 * sigma[0] ** 2)) + (Y ** 2 / (2 * sigma[1] ** 2))
              + (Z ** 2 / (2 * sigma[2] ** 2))))
        return f

    def rebin(self, factor):
        """
        Rebins the dataset with given factor which contains three uints
        indicating the binning factor for each dimension. Rebinning is
        performed by convoluting the dataset with a tophat function of given
        size. This makes that some edge artifacts will remain.

        Parameters
        ----------
        factor : [a1, a2, a3]
            The three binning factors. The values of a1, a2 and a3 >= 1 and
            integer.

        Returns
        -------
        None.
        """
        # todo use fftconvolve and zero padding

        if (factor[0] == 1) & (factor[1] == 1) & (factor[2] == 1):
            return self

        top_hat = np.ones(factor)
        top_hat = top_hat / top_hat.sum()
        # note that this is emperical found and I do not know
        # if this is true for everything
        sind = (np.array(factor) - 1) // 2

        res = ndimage.convolve(self.multidata, top_hat, mode='constant',
                               cval=0.0)

        ndat = res[sind[0]::factor[0], sind[1]::factor[1], sind[2]::factor[2]]
        spc = MultiSpectrumshape(factor[2] * self.dispersion,
                                 self.offset + factor[2] * self.dispersion,
                                 ndat.shape[2], ndat.shape[0], ndat.shape[1])
        s = MultiSpectrum(spc, data=ndat)
        return s

    def gaussiansmooth(self, sigma, crop=False):
        """
        Uses the 3D FFT to compute a gaussian blurred EELS map.
        Note the this performs a 3d fft which makes that the computation needs
        a lot of RAM

        Parameters
        ----------
        sigma : [a1, a2, a3]
            The sigma value of the 3d gaussian.

        crop: boolean
            Indicates whether the filtered spectrum will be

        Returns
        -------
        cop: MultiSpectrum
            The gaussian smoothed multispectrum
        """

        fft_sig = np.pi ** 2 / np.array(sigma)  # The sigma for FFT Gaussian

        xax = np.arange(self.xsize) - self.xsize / 2
        yax = np.arange(self.ysize) - self.ysize / 2
        X, Y, E = np.meshgrid(yax, xax, self.energy_axis
                              - self.energy_axis[int(self.size / 2)])

        FFT_d = np.fft.fftshift(np.fft.fftn(self.multidata))
        gauss = self.gaussian_3d(X, Y, E, fft_sig)
        res = np.real(np.fft.ifftn(np.fft.ifftshift(FFT_d * gauss)))

        if crop:
            # TODO better implementation of cropping can be thought off
            crop = [int(np.ceil(4 * sig)) for sig in sigma]
            crop[2] = int(np.ceil(crop[2] / self.dispersion))
            ndata = res[crop[0]:-crop[0], crop[1]:-crop[1], crop[2]:-crop[2]]
            noffset = crop[2] * self.dispersion + self.offset
            ms = MultiSpectrumshape(self.dispersion, noffset, ndata.shape[2],
                                    ndata.shape[0], ndata.shape[1])
            cop = MultiSpectrum(ms, data=ndata)
        else:
            cop = self.copy()
            cop.multidata = res

        return cop

    def firstderivative(self):
        """
        Calculates the first derivative of the multispectrum with respect to
        the energy axis. At this point I do not see any reason on implementing
        this for the scanning dimensions.

        Returns
        -------
        s: MultiSpectrum
            The first derivative for the multispectrum
        """

        differential = np.zeros_like(self.multidata)
        differential[:, :, 0] = self.multidata[:, :, 1] \
            - self.multidata[:, :, 0]
        differential[:, :, -1] = self.multidata[:, :, -1] \
            - self.multidata[:, :, -2]
        differential[:, :, 1:-1] = .5 * self.multidata[:, :, 2:] \
            - .5 * self.multidata[:, :, 0:-2]
        differential /= self.dispersion

        nspec = MultiSpectrumshape(self.dispersion, self.offset, self.size,
                                   self.xsize, self.ysize)

        s = MultiSpectrum(nspec, data=differential)
        return s

    def secondderivative(self):
        """
        Calculates the second derivative of the multispectrum with respect to
        the energy axis.

        Returns
        -------
        s: MultiSpectrum
            The second derivative for the multispectrum
        """
        differential = np.zeros_like(self.multidata)
        differential[:, :, 0] = self.multidata[:, :, 0] \
            - 2. * self.multidata[:, :, 1] + self.multidata[:, :, 2]
        differential[:, :, -1] = self.multidata[:, :, -1] - 2. \
            * self.multidata[:, :, -2] + self.multidata[:, :, -3]
        differential[:, :, 1:-1] = self.multidata[:, :, 0:-2] - 2. \
            * self.multidata[:, :, 1:-1] + self.multidata[:, :, 2:]

        differential /= self.dispersion ** 2

        nspec = MultiSpectrumshape(self.dispersion, self.offset,
                                   differential.shape[2], self.xsize,
                                   self.ysize)
        s = MultiSpectrum(nspec, data=differential)
        return s

    def upsample(self, nn):
        """
        Upsamples the spectrum with factor nn.

        Parameters
        ----------
        nn : uint
            The number of times the spectrum needs to be upsampled

        Returns
        -------
        spec_up: MultiSpectrum
            The upsampled multispectrum


        """
        shape = (self.xsize, self.ysize)
        spc = Spectrumshape(self.dispersion / nn, self.offset,
                            nn * self.size)
        s = Spectrum(spc)
        up = np.zeros(shape + (spc.size,))

        for index in tqdm(np.ndindex(shape),total=np.prod(shape),leave=True,position=0):
            islice = np.s_[index]

            self.setcurrentspectrum(islice)
            int_spec = self.interp_to_other_energy_axis(s)
            up[islice] = int_spec.data
        spec_up = MultiSpectrum.from_numpy(up, int_spec.energy_axis)

        return spec_up

    def map_to_line(self):
        """
        Make a map into a line scan
        :return:
        """
        new_shape = (int(self.xsize * self.ysize), 1, self.size)
        reshaped = np.reshape(self.multidata, new_shape)
        nspec = MultiSpectrumshape(self.dispersion, self.offset, self.size,
                                   new_shape[0], 1)
        s = MultiSpectrum(nspec, data=reshaped)
        return s

    def to_logarithm(self):
        s = self.copy()
        s.multidata[s.multidata < 1] = 1
        s.multidata = np.log(s.multidata)
        return s

    def apply_dark_and_gain(self, gain_name, dark_name):
        s = self.apply_dark(dark_name)
        gain = np.load(gain_name, allow_pickle=True)
        s.multidata = gain * s.multidata

        return s

    def apply_dark(self, dark_name):
        s = self.copy()
        dark = np.load(dark_name, allow_pickle=True)
        s.multidata -= dark
        return s

    def save_hdf5(self, filename, metadata=None, overwrite=False):
        file_exists = exists(filename)
        if not overwrite and file_exists:
            logger.info(
                r'File already exists, set overwrite to True '
                r'if you want to overwrite existing file')
            print('not overwriting')
            return False

        if file_exists:
            os.remove(filename)

        f = h5py.File(filename, 'w')
        f.create_dataset('Data', data=self.multidata)
        d = f.create_group('Energy_axis')

        d.attrs['Dispersion'] = self.dispersion
        d.attrs['Offset'] = self.offset
        d.attrs['Size'] = self.size

        if metadata is not None:
            print('saving metadata')
            if 'alpha' in metadata:
                d.attrs['alpha'] = metadata['alpha']
            if 'beta' in metadata:
                d.attrs['beta'] = metadata['beta']
            if 'E0' in metadata:
                d.attrs['E0'] = metadata['E0']
            if 'elements' in metadata:
                d.attrs['elements'] = metadata['elements']
            if 'edges' in metadata:
                d.attrs['edges'] = metadata['edges']
            if 'Layers' in metadata:
                print('saving layers')
                f.create_dataset('Layers', data=metadata['Layers'])
            if 'Layernames' in metadata:
                print('saving layers')
                f.create_dataset('Layernames', data=metadata['Layernames'])
            if 'Layerlow' in metadata:
                print('saving layers')
                f.create_dataset('Layerlow', data=metadata['Layerlow'])
            if 'Layerhigh' in metadata:
                print('saving layers')
                f.create_dataset('Layerhigh', data=metadata['Layerhigh'])
            if 'Layervisible' in metadata:
                print('saving layers')
                f.create_dataset('Layervisible', data=metadata['Layervisible'])

        f.close()
        return True



from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge


class DummyEdge(CoreLossEdge):
    """
    DummyEdge is a first approximation of the edge. This assumes each edge has
    a different onset energy but the powerlaw value is can be chosen. Note that
    the parameters E0, alpha, beta do not influence the dummy edge model.
    """

    def __init__(self, specshape, A, E0=300e3, alpha=1e-9, beta=10e-3,
                 element=None, edge=None, eshift=0, r=3):
        """

        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        A : float
            Amplitude of the edge

        E0: float [V]
            The acceleration voltage of the microscope

        alpha: float [rad]
            The convergence angle of the incoming probe

        beta: float [rad]
            The collection angle

        element: string
            The name of the element from which to calculate the edge model.

        edge: string
            The type of edge. (K1, L1, L2, L3, M1, etc.)

        eshift: float [eV]
            The shift of the onset energy with respect to the literature value.
            (default: 0)
        r: float
            The powerlaw of the dummyedge. (default: 3)

        Returns
        -------
        """

        super().__init__(specshape, A, E0, alpha, beta, element, edge,
                         eshift=eshift)
        self.r = r

    @classmethod
    def edge_by_onset(self, specshape, A, onset):
        """
        Class method to make a dummy edge where the edge onset energy is
        specified. In the list of elements not every edge is included so it
        would be of interest to have some freedom if needed.

        :param specshape:
        :param A:
        :param onset:
        :return:
        """
        d = DummyEdge(specshape, A, 1, 1, 1, 'C', 'K')
        d.onset_energy = onset
        return d

    def _check_str_in_list(self, list, edge):
        for name in list:
            if name[0] == edge[0]:
                return True

        return False

    def set_edge(self, edge):
        """
        Checks if the given edge is valid and adds the directories of the
        :param edge:
        :return:
        """
        edge_list = ['K1', 'L1', 'L2', 'L3', 'M2', 'M3', 'M4', 'M5', 'N4',
                     'N5']

        elem_list = self.get_elements()

        if not isinstance(edge, str):
            raise TypeError(
                'Edge should be a string: K1, L1, L2, L3, M2, M3, M4, M5, N4,'
                ' N5')
        if edge in edge_list:
            if self._check_str_in_list(elem_list, edge):
                self.edge = edge
            else:
                raise ValueError(r'The element {} does not'
                                 r' have a {} edge'.format(self.element, edge))

        else:
            raise ValueError(
                'Edge should be: K1, L1, L2, L3, M2, M3, M4, M5, N4, N5')

    def calculate_cross_section(self):
        cross_section = self.prefactor * self.energy_axis ** (-self.r)
        scale = 1 / cross_section[self.get_energy_index(self.onset_energy)]
        cross_section = cross_section * scale
        boolean = self.energy_axis < self.onset_energy
        cross_section[boolean] = 0
        return cross_section



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np
import logging

logger = logging.getLogger(__name__)


class PowerLaw(Component):
    """
    A power law background model

    Parameters
    ----------
    specshape : Spectrumshape
        The spectrum shape used to model

    A : float
        Amplitude of the powerlaw. Defined such that the offset
        energy has the value of the amplitude A.

    r: float
        The exponent value of the powerlaw.

    Returns
    -------
    """

    def __init__(self, specshape, A, r):
        super().__init__(specshape)
        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(0, 1e20)
        p1.sethasgradient(True)
        self._addparameter(p1)

        p2 = Parameter('r', r)
        p2.setlinear(False)
        p2.setboundaries(0.0, 1e20)
        p2.sethasgradient(True)
        self._addparameter(p2)
        # don't convolute the background it only gives problems and
        # adds no extra physics
        self._setcanconvolute(False)
        self.Estart = specshape.offset

        self._setname('Powerlaw')

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        if p1.ischanged() or p2.ischanged():
            A = p1.getvalue()
            r = p2.getvalue()
            self.data = self.powerlaw(A, r)
        self.setunchanged()  # put parameters to unchanged

    def powerlaw(self, A, r):
        return A * (self.energy_axis / self.Estart) ** (-r)

    def getgradient(self, parameter):
        """calculate the analytical partial derivative wrt parameter
        returns a reference to the gradient
        """
        if not parameter.hasgradient:
            return None
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        A = p1.getvalue()
        r = p2.getvalue()
        E = self.energy_axis
        mask = (E > 0) * (r > 0)  # only meaningfull when E>0 and r>0
        if parameter == p1:
            self.gradient[0] = mask * self.powerlaw(1, r)
            return self.gradient[0]
        elif parameter == p2:
            self.gradient[1] = mask * self.powerlaw(A, r) * np.log(
                self.Estart / E)

            return self.gradient[1]
        else:
            # throw Componenterr::bad_index()
            return None

    def autofit_(self, spectrum, istart, istop):
        """
        Performs a first guess of the parameters from the Powerlaw fit.
        It checks if the obtained values are reasonable and will apply
        them to the components parameters. The procedure is explained
        in Electron Energy Loss Spectroscopy (Third Edition) of Egerton
        in ...

        Parameters
        ----------
        spectrum : Spectrum
            The spectrum data from which an estimate is done on the background
        istart: int
            Index from where to start
        istop: int
            Index indicating the end of the fitting region

        Returns
        -------
        None.

        """

        # powerlaw fit to autodetermine the A and r parameter.
        if istart >= istop:
            # impossible to fit
            A = 0.0
            r = 3.0
            return
        N = istop - istart
        E = self.energy_axis

        # added to remove pixel values smaller than 1 since they give a
        # problem with the log
        ndat = np.copy(spectrum.data[istart:istop])
        # ndat[ndat<1.] = 1
        boolean = ndat >= 1

        x = np.log(E[istart:istop][boolean])
        N = x.size
        #

        y = np.log(ndat[boolean])
        xy = (x * y).sum()
        sx = x.sum()
        sy = y.sum()
        sxsq = (x ** 2).sum()

        b = (N * xy - sx * sy) / (N * sxsq - sx ** 2.0)
        a = sy / N - b * sx / N
        r = -b
        A = np.exp(a) * E[0] ** -r  # correct for referencing A to E[0]

        # print(A)
        # print(r)

        # todo modified the checks for the fast background measurement
        # sanity check
        if (r < 0.0) or (r > 10):
            r = 3.0
            A = spectrum.data[0]
            logger.warning('Powerlaw autofit resulted in r<1, ignoring output')
        if (A < 0.0):
            r = 3.0
            A = spectrum.data[0]
            logger.warning('Powerlaw autofit resulted in A<0, ignoring output')

        # logger.info('PowerLaw autofit resulted in A=%e, r=%e',A,r)
        # set the parameters
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p1.setvalue(A)
        p2.setvalue(r)
        self.calculate()

    def autofit(self, spectrum, istart, istop):
        """
        Performs a first guess of the parameters from the Powerlaw fit.
        It checks if the obtained values are resonable and will apply
        them to the components parameters. The procedure is explained
        in Electron Energy Loss Spectroscopy (Third Edition) of Egerton
        in ...

        Parameters
        ----------
        spectrum : Spectrum
            The spectrum data from which an estimate is done on the background
        istart: int
            Index from where to start
        istop: int
            Index indicating the end of the fitting region

        Returns
        -------
        None.

        """

        # powerlaw fit to autodetermine the A and r parameter.
        if istart >= istop:
            # impossible to fit
            A = 0.0
            r = 3.0
            return
        N = istop - istart
        E = self.energy_axis

        # added to remove pixel values smaller than 1 since they give a
        # problem with the log
        ndat = np.copy(spectrum.data[istart:istop])
        # ndat[ndat<1.] = 1

        boolean = ndat > 1.
        x = np.log(E[istart:istop][boolean])
        N = x.size

        y = np.log(ndat[boolean])
        sx = x.sum()

        D = np.nansum(x ** 2 / y) * np.nansum(1 / y) - np.nansum(x / y) ** 2
        r = -1 * (sx * np.nansum(1 / y) - N * np.nansum(x / y)) / D
        logA = (N * np.nansum(x ** 2 / y) - sx * np.nansum(x / y)) / D
        A = np.exp(logA) * E[0] ** -r

        # print('A: ' + str(A))
        # print('r: ' + str(r))

        # todo modified the checks for the fast background measurement
        # sanity check
        # if (r < 0.0) or (r > 10):
        #     r = 3.0
        #     A = spectrum.data[0]
        #     logger.warning('Powerlaw autofit resulted in r<1, 3
        #     ignoring output')
        # if (A < 0.0):
        #     r = 3.0
        #     A = spectrum.data[0]
        #     logger.warning('Powerlaw autofit resulted in A<0,
        #     ignoring output')

        # logger.info('PowerLaw autofit resulted in A=%e, r=%e',A,r)
        # set the parameters
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        b1 = p1.setvalue(A)
        b2 = p2.setvalue(r)
        self.calculate()
        return b1 and b2

    def autofit1(self, spectrum, istart, istop):
        """
        Performs a first guess of the parameters from the Powerlaw fit.
        It checks if the obtained values are resonable and will apply
        them to the components parameters. The procedure is explained
        in Electron Energy Loss Spectroscopy (Third Edition) of Egerton
        in ...

        Parameters
        ----------
        spectrum : Spectrum
            The spectrum data from which an estimate is done on the background
        istart: int
            Index from where to start
        istop: int
            Index indicating the end of the fitting region

        Returns
        -------
        None.

        """

        # powerlaw fit to autodetermine the A and r parameter.
        if istart >= istop:
            # impossible to fit
            A = 0.0
            r = 3.0
            return
        E = self.energy_axis

        x = np.log(E[istart:istop])

        ndat = np.copy(spectrum.data[istart:istop])
        ndat[ndat < 1.] = 1

        # w = 1/np.sqrt(ndat)
        # todo have a proper description of the weights for linear fitter
        # w = 1/ndat
        w = np.ones(ndat.size)

        y = np.log(ndat)
        xw = (w * x).sum() / w.sum()
        yw = (w * y).sum() / w.sum()
        b = np.sum(w * (x - xw) * (y - yw)) / np.sum(w * (x - xw) ** 2)
        a = yw - b * xw
        r = -b
        A = np.exp(a) * E[0] ** -r  # correct for referencing A to E[0]

        # todo modified the checks for the fast background measurement
        # sanity check
        # if (r < 0.0) or (r > 10):
        #     r = 3.0
        #     A = spectrum.data[0]
        #     logger.warning('Powerlaw autofit resulted in r<1,
        #     ignoring output')
        # if (A < 0.0):
        #     r = 3.0
        #     A = spectrum.data[0]
        #     logger.warning('Powerlaw autofit resulted in A<0,
        #     ignoring output')

        # logger.info('PowerLaw autofit resulted in A=%e, r=%e',A,r)
        # set the parameters
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p1.setvalue(A)
        p2.setvalue(r)
        self.calculate()

    def autofit_matrix(self, spectrum, istart, istop):
        """
        Weighted linear least squares where the variance is given by
         the square root of the
        signal

        :param spectrum:
        :param istart:
        :param istop:
        :return:
        """
        if istart > istop:
            return

        subdata = spectrum.data[istart:istop]
        boolean = subdata >= 1.
        ln_y = np.log(subdata[boolean])
        W = np.diag(np.log(1 / (subdata[boolean])))
        A = np.zeros((boolean.sum(), 2))
        A[:, 0] = np.ones(boolean.sum())
        A[:, 1] = -np.log(
            self.energy_axis[istart:istop][boolean] / self.energy_axis[0])

        term1 = np.linalg.inv(np.dot(A.T, np.dot(W, A)))
        term2 = np.dot(A.T, np.dot(W, ln_y))
        x_ = np.dot(term1, term2)

        # x_, resi, rank, sing = np.linalg.lstsq(A, ln_y, rcond=-1)
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p1.setvalue(np.exp(x_[0]))
        p2.setvalue(x_[1])
        self.calculate()

    def autofit_twowindow(self, spectrum, istart, istop):
        """
        Performs a first guess of the parameters from the Powerlaw fit.
        It checks if the obtained values are resonable and will apply
        them to the components parameters. The procedure is explained
        in Electron Energy Loss Spectroscopy (Third Edition) of Egerton
        in ...

        This is the two-window technique. It is quick and dirty,
        and not as accurate as a proper least squares fit,
        but is has its uses.

        Parameters
        ----------
        spectrum : Spectrum
            The spectrum data from which an estimate is done on the background
        istart: int
            Index from where to start
        istop: int
            Index indicating the end of the fitting region

        Returns
        -------
        None.

        """

        # powerlaw fit to autodetermine the A and r parameter.
        if istart >= istop:
            # impossible to fit
            A = 0.0
            r = 3.0
            return

        y = np.copy(spectrum.data[istart:istop])  # istop = non-inclusive

        x0 = self.energy_axis[istart]
        x1 = self.energy_axis[istop - 1]  # istop - 1 is inclusive

        I0 = np.sum(y[:len(y) // 2])
        I1 = np.sum(y[len(y) // 2:])
        r = 2. * np.log(I0 / I1) / np.log(
            x1 / x0)  # the usual power-law exponent
        A = (I0 + I1) * (x0 * x1) ** (r / 2.) / (x1 - x0)
        A *= self.Estart ** (-r)

        # todo modified the checks for the fast background measurement
        # sanity check
        if (r < 0.0) or (r > 10):
            r = 3.0
            A = spectrum.data[0]
            logger.warning('Powerlaw autofit resulted in r<1, ignoring output')
        if (A < 0.0):
            r = 3.0
            A = spectrum.data[0]
            logger.warning('Powerlaw autofit resulted in A<0, ignoring output')

        # logger.info('PowerLaw autofit resulted in A=%e, r=%e',A,r)
        # set the parameters
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p1.setvalue(A)
        p2.setvalue(r)
        self.calculate()

    def autofit_areas(self, spectrum, istart, istop):
        """
        Performs a first guess of the parameters from the Powerlaw fit.
        It checks if the obtained values are resonable and will apply
        them to the components parameters. The procedure is explained
        in Electron Energy Loss Spectroscopy (Third Edition) of Egerton
        in Chapter 4.4.2. (ISBN: 978-1-4419-9583-4).

        Parameters
        ----------
        spectrum : Spectrum
            The spectrum data from which an estimate is done on the background
        istart: int
            Index from where to start
        istop: int
            Index indicating the end of the fitting region

        Returns
        -------
        None.

        """
        boolean = (istart + istop) % 2 == 0
        if not boolean:
            istop -= 1

        if istart - istop == 0:
            istop = istart + 2

        if istart >= istop:
            # impossible to fit
            A = 0.0
            r = 3.0
            return

        imiddle = (istop + istart) // 2
        I1 = spectrum.data[istart:imiddle].sum()
        I2 = spectrum.data[imiddle:istop].sum()
        E = spectrum.energy_axis
        r = 2 * np.log(I1 / I2) / (np.log(E[istop] / E[istart]))
        k = 1 - r
        # Since E0 is used there is another normalization
        # in this case it is found experimentally instead of proven
        # (so can be wrong)
        A = (k * I2 / ((E[istop] / E[0]) ** k - (E[imiddle] / E[0]) ** k)) / E[
            0]
        # sanity check
        if (r < 1.0) or (r > 10):
            r = 3.0
            A = spectrum.data[0]
            logger.warning('Powerlaw autofit resulted in r<1, ignoring output')
        if (A < 0.0):
            r = 3.0
            A = spectrum.data[0]
            logger.warning('Powerlaw autofit resulted in A<0, ignoring output')

        # logger.info('PowerLaw autofit resulted in A=%e, r=%e',A,r)
        # set the parameters
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p1.setvalue(A)
        p2.setvalue(r)
        self.calculate()



from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge
from pyEELSMODEL.components.CLedge.dummymodel import DummyEdge
import numpy as np
import logging

logger = logging.getLogger(__name__)


# todo not needed for

class DummyEdgeCombined(CoreLossEdge):
    """
    DummyEdge is a first approximation of the edge. This assumes each edge has
    a different onset energy but the powerlaw value is -3.
    Note that the parameters E0, alpha, beta do not influence the
    dummyedge model.

    """

    def __init__(self, specshape, A, E0, alpha, beta, element, edge, eshift=0):
        """
        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        A : float
            Amplitude of the edge

        E0: float [V]
            The acceleration voltage of the microscope

        alpha: float [rad]
            The convergence angle of the incoming probe

        beta: float [rad]
            The collection angle

        element: string
            The name of the element from which to calculate the edge model.

        edge: string
            The type of edge. (K, L or M)

        eshift: float [eV]
            The shift of the onset energy with respect to the literature value.
            (default: 0)


        Returns
        -------
        """

        self.xsectionlist = []
        if edge == 'K':
            xsectionK = DummyEdge(specshape, A, E0, alpha, beta, element, 'K')
            self.xsectionlist.append(xsectionK)
            super().__init__(specshape, A, E0, alpha, beta, element, 'K')
            name = element + ' K edge: ' + str(self.onset_energy) + ' eV'
            self.setdisplayname(name)
        if edge == 'L':

            xsectionL3 = DummyEdge(specshape, A, E0, alpha, beta, element,
                                   'L3')
            # xsectionL3.prefactor = 2/3
            try:
                xsectionL2 = DummyEdge(specshape, A, E0, alpha, beta, element,
                                       'L2')
                # xsectionL2.prefactor = 1 / 3
            except Exception:
                print('No L2 is given which means it is the same as L3')
                xsectionL2 = DummyEdge(specshape, A, E0, alpha, beta, element,
                                       'L3')
                xsectionL2.prefactor = 1 / 2

            xsectionL2.parameters[0].couple(xsectionL3.parameters[0])
            xsectionL2.parameters[1].couple(xsectionL3.parameters[1])
            xsectionL2.parameters[2].couple(xsectionL3.parameters[2])
            xsectionL2.parameters[3].couple(xsectionL3.parameters[3])

            self.xsectionlist.append(xsectionL3)
            self.xsectionlist.append(xsectionL2)
            super().__init__(specshape, A, E0, alpha, beta, element, 'L3',
                             eshift=eshift, q_steps=100)
            name = element + 'L3,2,1 edge: ' + str(self.onset_energy) + ' eV'
            self.setdisplayname(name)

        elif edge == 'M':
            xsectionM5 = DummyEdge(specshape, A, E0, alpha, beta, element,
                                   'M5')
            xsectionM4 = DummyEdge(specshape, A, E0, alpha, beta, element,
                                   'M4')

            xsectionM4.parameters[0].couple(xsectionM5.parameters[0])
            xsectionM4.parameters[1].couple(xsectionM5.parameters[1])
            xsectionM4.parameters[2].couple(xsectionM5.parameters[2])
            xsectionM4.parameters[3].couple(xsectionM5.parameters[3])

            self.xsectionlist.append(xsectionM5)
            self.xsectionlist.append(xsectionM4)

            super().__init__(specshape, A, E0, alpha, beta, element, 'M5',
                             eshift=eshift, q_steps=100)
            name = element + 'M5,4,3,2 edge' + str(self.onset_energy) + ' eV'
            self.setdisplayname(name)

        elif edge == 'N':
            logger.error('N edge not implemented yet')

        self.manageparameters()
        self.calculate()

    def manageparameters(self):
        # erase existing parameter list and replace with longer list
        self.parameters = []
        self.gradient = []
        # first 4 paramters however remain same meaning but they are
        # coupled with the higher ones
        for xsection in self.xsectionlist:
            for par in xsection.parameters:
                self._addparameter(par)

    def set_onset_energy(self):
        self.onset_energy = self.xsectionlist[
            0].onset_energy  # first part of cross section determines onset

    def set_edge(self, edge):
        self.edge = self.xsectionlist[0].edge

    def calculate_cross_section(self):
        cross_section = np.zeros(self.size)
        for xsection in self.xsectionlist:
            if xsection.suppress==False:
                cross_section += xsection.calculate_cross_section()
            else:
                print('subcomponent is suppressed, to unset call setsuppress(False) on the specific subcomponent:',xsection.name)            
        return cross_section

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        pA = self.parameters[0]
        p2 = self.parameters[1]
        p3 = self.parameters[2]
        p4 = self.parameters[3]

        if p2.ischanged() or p3.ischanged() or p4.ischanged():
            self.cross_section = self.calculate_cross_section()
            self.data = pA.getvalue() * self.cross_section

        if pA.ischanged():
            self.data = pA.getvalue() * self.cross_section
        self.setunchanged()



from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape
from pyEELSMODEL.core.spectrum import Spectrum, Spectrumshape
import numpy as np


class Mscatter(Component):
    """
    A convolutor component can take the input of another component and
    apply a convolution operator to it and store the results of that operation
    in its own spectrum data. The model will treat this as special and first
    calculate the individual components that have canconvolute set, sum those
    and then apply these as input to a convolutor component (typically
    only 1 in a model). After this those components are added which have
    canconvolute=False (eg a background component)
    """

    def __init__(self, specshape, llspectrum):
        """

        Parameters
        ---------
        specshape: Spectrumshape
            The spectrum shape on the spectrum it will be used, not the low
            loss spectrum
        llspectrum: Spectrum or MultiSpectrum
            The spectrum or multispectrum which is used to convolve the rest
            of the components with.

        """
        super().__init__(specshape)

        n_acc = 8  # sometimes round error make that disperion is not the same
        if np.round(self.dispersion, n_acc) != np.round(llspectrum.dispersion,
                                                        n_acc):
            raise ValueError('low loss spectrum has to have same '
                             'dispersion as model')

        if self.size < llspectrum.size:
            raise ValueError('size of ll spectrum is larger than the spectrum'
                             ' on which the analysis will be performed')

        elif self.size > llspectrum.size:
            # print('low loss spectrum has to have same size as model')
            # print('the low loss will be zero padded to have the same size')
            llspectrum = self.padding(specshape, llspectrum)

        self._setcanconvolute(False)  # meaningless in this case
        self._setshifter(False)
        self._isconvolutor = True
        self.llspectrum = llspectrum

    def padding(self, specshape, llspectrum):
        """
        Zero pads the low loss spectrum to have the same size as the spectrum
        shape.
        Parameters
        ---------
        specshape: Spectrumshape
            The spectrum shape on the spectrum it will be used, not the low
            loss spectrum
        llspectrum: Spectrum or MultiSpectrum
            The spectrum or multispectrum which is used to convolve the rest
            of the components with.
        """
        size_dif = specshape.size - llspectrum.size
        before = size_dif // 2
        after = size_dif - before
        # print('padding is done')
        # print(llspectrum)
        if type(llspectrum) is Spectrum:
            # print('low loss is spectrum')
            pad_data = np.pad(llspectrum.data, pad_width=(before, after))
            noffset = llspectrum.offset - llspectrum.dispersion * before
            sph = Spectrumshape(llspectrum.dispersion, noffset, pad_data.size)
            s = Spectrum(sph, data=pad_data)

        elif type(llspectrum) is MultiSpectrum:
            # print('low loss is multispectrum')
            pad_data = np.pad(llspectrum.multidata,
                              pad_width=((0, 0), (0, 0), (before, after)))
            noffset = llspectrum.offset - llspectrum.dispersion * before
            sph = MultiSpectrumshape(llspectrum.dispersion, noffset,
                                     pad_data.shape[-1],
                                     llspectrum.xsize, llspectrum.ysize)
            s = MultiSpectrum(sph, data=pad_data)

        return s

    def save(self, fh):
        # save the details of this component to file in such a way
        # that you can bring it to the same state later

        # save the file location of the ll spectrum to open it later
        # when reloading the model

        return

    def load(self, fh):
        # load this component from saved detaile

        # get filename
        # open the ll spectrum if not already open

        # self.llspectrum=...
        return

    def calculate(self):
        print(
            'if this statement is printed, it means that no '
            'calculate function exists for other MScatter components')



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  9 20:46:09 2021

@author: joverbee
"""
import numpy as np
import matplotlib.pyplot as plt
import logging
from pyEELSMODEL.operators.deconvolutions.deconvolution import Deconvolution
from pyEELSMODEL.components.gaussian import Gaussian
from pyEELSMODEL.core.multispectrum import MultiSpectrum
from pyEELSMODEL.core.spectrum import Spectrum
from tqdm import tqdm

logger = logging.getLogger(__name__)


def fft_phys(data):
    return np.fft.fftshift(np.fft.fft(np.fft.ifftshift(data)))


def ifft_phys(data):
    return np.fft.ifftshift(np.fft.ifft(np.fft.fftshift(data)))


class GaussianModifier(Deconvolution):
    """
    The gaussian modifier deconvolution as described in
    10.1016/j.ultramic.2009.06.010.
    """
    def __init__(self, spectrum, llspectrum, factor=None):
        """
        Parameters
        ----------
        spectrum: Spectrum or MultiSpectrum
            The spectrum which needs to be deconvolved. Best practice is to
            remove the background before deconvolving.
        llspectrum: Spectrum or MultiSpectrum
            The low loss which is used for the deconvolution
        factor: uint
            This determines the width of the gaussian used. The fhwm
            of the gaussian is factor * dispersion. If factor is None, the
            2*fwhm of the zero-loss peak is used.

        """
        super().__init__(spectrum, llspectrum)
        self.restored = None
        self.restored = None

        if factor is None:
            self.factor = 2 * llspectrum.get_numerical_fwhm() \
                          / llspectrum.dispersion
        else:
            self.factor = factor
        # debugging feature and to optimize gaussian filter
        self.plotting = False

    def gaussianmodifier(self, signal, psf, centre, fwhm, plotting=False):
        """
        The gaussian modfier algorithm.
        """
        g = Gaussian(self.spectrum.get_spectrumshape(), A=1, centre=centre,
                     fwhm=fwhm)
        g.calculate()

        gdata = g.data / g.data.sum()

        co = np.argmax(psf)
        rol_co = int(psf.size / 2) - co

        if plotting:
            plt.figure()
            plt.plot(np.abs(np.fft.rfft(gdata)), label='FFT Gaussian')
            plt.plot(np.abs(np.fft.rfft(psf)), label='FFT Low Loss')
            plt.legend()

        O_f = np.fft.rfft(gdata) * np.fft.rfft(signal) / np.fft.rfft(psf)

        O_E = np.roll((np.fft.irfft(O_f)), -rol_co)

        return O_E

    def multi_deconvolve(self):
        """
        Performs the deconvolution on the  multispectrum
        """

        shape = (self.spectrum.xsize, self.spectrum.ysize)
        ms = self.spectrum.copy()
        centre = self.spectrum.energy_axis[int(self.spectrum.size/2)]
        sigma = self.spectrum.dispersion * self.factor

        for index in tqdm(np.ndindex(shape)):
            islice = np.s_[index]
            self.spectrum.setcurrentspectrum(index)
            self.llspectrum.setcurrentspectrum(index)
            psf = self.llspectrum.data / self.llspectrum.data.sum()

            restore = self.gaussianmodifier(self.spectrum.data, psf, centre,
                                            sigma, plotting=self.plotting)
            ms.multidata[islice] = restore

        return ms

    def deconvolve(self):
        centre = self.spectrum.energy_axis[int(self.spectrum.size/2)]
        fwhm = self.spectrum.dispersion*self.factor
        psf = self.llspectrum.data/self.llspectrum.data.sum()

        if type(self.spectrum) is MultiSpectrum:
            s = self.multi_deconvolve()
            self.restored = s
        else:
            restore = self.gaussianmodifier(self.spectrum.data, psf, centre,
                                            fwhm, plotting=self.plotting)
            s = Spectrum(self.spectrum.get_spectrumshape(), data=restore)
            self.restored = s
        return s



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 29 16:25:55 2021

@author: joverbee
"""

import sys
from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge
from pyEELSMODEL.core.spectrum import Spectrumshape
import pytest
sys.path.append("..")  # Adds higher directory to python modules path.


def test_set_element():
    specshape = Spectrumshape(1, 100, 1024)
    OK = CoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3, beta=20e-3,
                      element='O', edge='K1')
    assert OK.element == 'O'

    with pytest.raises(ValueError):
        CoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3, beta=20e-3,
                     element='OK', edge='K1')


def test_set_Z():
    specshape = Spectrumshape(1, 100, 1024)
    OK = CoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3, beta=20e-3,
                      element='O', edge='K1')
    CK = CoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3, beta=20e-3,
                      element='C', edge='K1')
    TiL = CoreLossEdge(specshape, A=1e3, E0=300e3, alpha=5e-3, beta=20e-3,
                       element='Ti', edge='L3')

    assert OK.Z == 8
    assert CK.Z == 6
    assert TiL.Z == 22


def main():
    test_set_element()
    test_set_Z()


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
import logging
from matplotlib.patches import Rectangle
from pyEELSMODEL.core.operator import Operator

logger = logging.getLogger(__name__)


class MultiSpectrumVisualizer(Operator):

    """
    Class which visualizes the multspectrum.
    """
    def __init__(self, multispectra, input_map=None, labels=None,
                 logscale=False):
        """
        Class which visualizes the multspectrum.
        Parameters
        ----------
        multispectra: list of MultiSpectrums
            A list of multispectrum
        input_map: 2d numpy array
            An 2d array which can be used to visualize the regions. For
            instance the ADF acquired simultaneously can be used. If None,
            then the sum of the spectrum at each probe positions will be
            used.
        labels: List of strings
            A label to indicate which spectrum is visualized with which color
        logscale: bool
            Indicates if the y axis should be visualized as a log scale
            (default: False)

        """
        self.multispectra = multispectra
        self.logscale = logscale
        self.is_line = False
        if self.multispectra[0].xsize == 1 or self.multispectra[0].ysize == 1:
            self.is_line = True
            input_map = np.squeeze(self.multispectra[0].multidata)
            self.w = self.multispectra[0].size

        if input_map is None:
            self.input_map = self.multispectra[0].multidata.sum(2)
        else:
            self.input_map = input_map

        self.sx = 0
        self.sy = 0
        self.h = 1
        self.w = 1
        self.linewidth = 5
        self.title = None

        # if we have a line scan, we auto scale the imshow
        if 1 in self.input_map.shape:
            self.aspect = 'auto'
        else:
            self.aspect = None
        self.press = None

        self.xlim_eels = None
        if labels is None:
            self.labels = [None]*len(self.multispectra)
        else:
            self.labels = labels

        self.fig, self.ax = plt.subplots(1, 2)
        self.plot()
        self.connect()

    def get_indextitle(self, x0, y0, w, h):
        title = 'x0, y0: ' + str(int(x0)) + ', '+str(int(y0)) + '; w, h: ' \
                + str(int(h)) + ', '+str(int(w))
        return title

    def plotrect(self):
        # plot selection rectangle on top of the image plot
        ax = self.ax
        if self.is_line:
            ax[0].axis("tight")  # fill window space
            xmin, xmax = ax[0].get_xlim()
            self.w = self.multispectra[0].size
            self.rect = Rectangle((self.sx - 0.5, self.sy - 0.5),
                                  self.w, self.h, fill=False, edgecolor='red',
                                  linewidth=self.linewidth)
        else:
            self.rect = Rectangle((self.sx - 0.5,  self.sy - 0.5),
                                  self.w, self.h, fill=False, edgecolor='red',
                                  linewidth=self.linewidth)

            ax[0].axis("image")  # square pixels

        ax[0].axis('off')
        ax[0].add_patch(self.rect)

    def updaterect(self):
        # print(self.get_indextitle(self.sx, self.sy, self.w, self.h))
        self.rect.set_xy((self.sx-0.5, self.sy - 0.5))
        self.rect.set_width(self.w)
        self.rect.set_height(self.h)
        self.rect.figure.canvas.draw()

    def plot(self, **kwargs):
        ax = self.ax
        ax[0].imshow(self.input_map, aspect=self.aspect)
        self.plotrect()
        # print(self.w)
        # print(self.h)
        self.plotline = []

        for spectra, label in zip(self.multispectra, self.labels):
            y = [self.sx, self.sx+self.h]
            x = [self.sy, self.sy+self.w]
            mydata = spectra.multidata[x[0]:x[1], y[0]:y[1], :].mean((0, 1))
            plotline = ax[1].plot(spectra.energy_axis, mydata, label=label)
            # print('initialize')
            self.plotline.append(plotline)
        ax[1].legend()
        ax[1].set_title(self.get_indextitle(self.sy, self.sx, self.h, self.w))
        self.xlim = ax[0].get_xlim()
        self.ylim = ax[0].get_ylim()

        if self.logscale:
            ax[1].set_ylim([1, None])
            ax[1].set_yscale('log')

        for key, value in kwargs.items():
            if key == 'xlim':
                ax[1].set_xlim(value)
                self.xlim_eels = value

    def connect(self):
        """Connect to all the events we need."""
        self.cidpress = self.rect.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.rect.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.rect.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

        self.keypress = self.rect.figure.canvas.mpl_connect('key_press_event',
                                                            self.on_key_press)

    def on_key_press(self, event):
        dx = 1
        dy = 1
        dw = 1
        dh = 1
        # print('key pressed')

        x0 = self.sx
        y0 = self.sy
        w = self.w
        h = self.h
        if event.key == 'up':
            self.sy = int(max(0, y0 - dy))

        if event.key == 'down':
            self.sy = int(min(self.ylim[0] + 0.5 - h, y0 + dy))

        if event.key == 'left' and not self.is_line:
            self.sx = int(max(0, x0 - dx))

        if event.key == 'right' and not self.is_line:
            self.sx = int(min(self.xlim[1] + 0.5 - w, x0 + dx))

        if event.key == '+':
            bolx = (x0 + w + dw >= self.xlim[1])
            boly = (y0 + h + dh >= self.ylim[0])

            if bolx and not boly and self.is_line is False:
                self.h = h + dh

            # if not(bolx) and self.is_line==False:
            elif boly and not bolx:
                self.w = w + dw

            elif bolx and boly:
                print('Cannot make square larger')

            else:
                self.w = w + dw
                self.h = h + dh

        if event.key == '-':
            self.w = max(1, w - dw)
            self.h = max(1, h - dh)

        # self.rect.set_x(self.sx)
        # self.rect.set_y(self.sy)

        self.update_eels()

    def on_press(self, event):
        """Check whether mouse is over us; if so, store some data."""
        if event.inaxes != self.rect.axes:
            return
        contains, attrd = self.rect.contains(event)
        if not contains:
            return
        # print('event contains', self.rect.xy)
        self.press = self.rect.xy, (event.xdata, event.ydata)

    def on_motion(self, event):
        """Move the rectangle if the mouse is over us."""
        if self.press is None or event.inaxes != self.rect.axes:
            return
        (x0, y0), (xpress, ypress) = self.press
        dx = event.xdata - xpress
        dy = event.ydata - ypress
        # print(f'x0={x0}, xpress={xpress}, event.xdata={event.xdata}, '
        #       f'dx={dx}, x0+dx={x0+dx}')
        shiftx = int(max(0,
                         min(self.xlim[1] - self.rect.get_width(), x0 + dx)))
        shifty = int(max(0,
                         min(self.ylim[0] - self.rect.get_height(), y0 + dy)))

        self.sx = shiftx
        self.sy = shifty
        self.update_eels()
        self.rect.figure.canvas.draw()

    def on_release(self, event):
        """Clear button press information."""
        self.press = None
        self.rect.figure.canvas.draw()

    def update_eels(self):
        self.updaterect()

        ax = self.ax
        # self.rect.figure.axes[1].cla()
        # y0 = max(0, y0 + 0.5)
        # x0 = max(0, x0 + 0.5)
        #
        # print(self.sx)
        # print(self.sy)
        # print(self.h)
        # print(self.w)
        miny = 0
        maxy = 0
        for plotline, spectra, label in zip(self.plotline,
                                            self.multispectra, self.labels):
            y = [self.sx, self.sx+self.w]
            x = [self.sy, self.sy+self.h]
            mydata = spectra.multidata[x[0]:x[1], y[0]:y[1], :].mean((0, 1))
            plotline[0].set_ydata(mydata)
            miny = min(mydata.min(), miny)
            maxy = max(mydata.max(), maxy)
            # print('hello')
        ax[1].set_title(self.get_indextitle(self.sy, self.sx, self.w, self.h))

        ax[1].set_ylim([miny, maxy])
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def disconnect(self):
        """Disconnect all callbacks."""
        self.rect.figure.canvas.mpl_disconnect(self.cidpress)
        self.rect.figure.canvas.mpl_disconnect(self.cidrelease)
        self.rect.figure.canvas.mpl_disconnect(self.cidmotion)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.path as mplPath
from matplotlib.backend_bases import MouseButton
import logging
from pyEELSMODEL.core.operator import Operator

logger = logging.getLogger(__name__)


class AreaSelection(Operator):

    """
    Class which uses user input to return an average area of a multispectrum.
    This only will not work for a line scan since chosing an area out of this
    is a bit easier.
    """
    def __init__(self, multispectrum, input_map=None, max_points=4,
                 other_spectra=[]):
        """
        Parameters
        ----------
        spectrum: Multispectrum
            The multispectrum form which an area need to be selected
        input_map: 2d numpy array
            The image used to select the proper region. This could be the HAADF
            signal or other calculated images. It is necessary that the scan
            size is the same as for the multispectrum. If input_map is None,
            the summed signal from the MultiSpectrum is used. (default: None)
        max_points: uint > 1
            The number of points used to determine the area. (default: 4)
        """
        self.multispectrum = multispectrum
        self.input_map = input_map
        self.max_points = max_points
        self.other_spectra = other_spectra
        self.avg_spectrum = None
        self.xcoords = None
        self.ycoords = None
        self.other_avg_spec = []

    @property
    def input_map(self):
        return self._input_map

    @input_map.setter
    def input_map(self, input_map):
        if input_map is None:
            im = self.multispectrum.integrate((0, -1), index_type=True)
            self._input_map = im
        else:
            self._input_map = input_map

    @property
    def max_points(self):
        return self._max_points

    @max_points.setter
    def max_points(self, max_points):
        if max_points < 2:
            raise ValueError(r'The maximum point should be larger than 1.')
        self._max_points = max_points

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, mask):
        self._mask = mask

    def calculate_mask_from_coords(self):
        """
        Calculates the mask used from the points selected.
        This mask can be accessed via the attribute self.mask.

        """
        marray = np.array([self.xcoords, self.ycoords]).transpose()
        poly_path = mplPath.Path(marray)

        shape = (self.multispectrum.xsize, self.multispectrum.ysize)
        mask = np.zeros(shape, dtype=bool)
        X, Y = np.meshgrid(np.arange(shape[1]), np.arange(shape[0]))
        for index in (np.ndindex(shape)):
            islice = np.s_[index]
            point = (X[islice], Y[islice])
            mask[islice] = poly_path.contains_point(point)
        self.mask = mask

    def get_mean_from_area(self):
        """
        Uses the user input area to calculate the average EEL spectrum.
        This function uses the self.xcoords and self.ycoords to calculate
        the area. Therefore, it is necessary to have some values for these
        attributes. User input on these coordinates can be obtained via the
        self.determine_input_area() function. The average spectra of the
        self.other_spectra attribute can be accessed via the
        self.other_avg_spec attribute.

        Returns
        -------
        s: Spectrum
            The average spectrum from the area.


        """

        self.calculate_mask_from_coords()

        ndata = self.multispectrum.multidata[self.mask, :].mean((0))
        # also the exclude and all these things are carried over.
        s = self.multispectrum.sum().copy()
        s.data = ndata
        self.avg_spectrum = s

        for spec in self.other_spectra:
            ndata = spec.multidata[self.mask, :].mean((0))
            sn = spec.sum().copy()
            sn.data = ndata
            self.other_avg_spec.append(sn)

        return s

    def show_area(self, **kwargs):
        """
        Shows the area which is selected.
        """
        fig, ax = plt.subplots()
        ax.imshow(self.input_map, **kwargs)
        ax.fill(self.xcoords, self.ycoords, alpha=0.5, color='green')

    def determine_input_area(self):
        """
        Shows an image (or plot)  which is given by input map. By right
        clicking on the image, an area can be selected where the number of
        corners is defined by the max_points parameter.
        This function sets the proper values for self.xcoords and self.ycoords.

        """

        def onclick(event):
            if event.button is MouseButton.LEFT:
                pass
            else:
                global ix, iy
                ix, iy = event.xdata, event.ydata
                xcoords.append(ix)
                ycoords.append(iy)

                ax.scatter(event.xdata, event.ydata, color='black')
                if len(xcoords) > 1:
                    ax.plot([event.xdata, xcoords[-2]],
                            [event.ydata, ycoords[-2]], color='black')
                if len(ycoords) == self.max_points:
                    ax.plot([event.xdata, xcoords[0]],
                            [event.ydata, ycoords[0]], color='black')
                    ax.fill(xcoords, ycoords, alpha=0.5, color='black')
                    print('the shape is drawn, this will be used to determine '
                          'the average spectrum')
                    self.xcoords = xcoords
                    self.ycoords = ycoords

                fig.canvas.draw()

                if len(xcoords) == self.max_points:
                    fig.canvas.mpl_disconnect(cid)

                return xcoords

        fig, ax = plt.subplots()
        ax.imshow(self.input_map)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.show()

        global xcoords, ycoords
        xcoords = []
        ycoords = []
        # self.xcoords = np.array(xcoords)
        # self.ycoords = np.array(ycoords)
        self.xcoords = xcoords
        self.ycoords = ycoords

        cid = fig.canvas.mpl_connect('button_press_event', onclick)



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from scipy import special
import numpy as np
from pyEELSMODEL.core.parameter import Parameter


class Voigt(Component):
    """
        Initialises a voigt component. The voigt function is the convolution
        of a gaussian with a lorentzian function.


    """
    def __init__(self, specshape, A, centre, gamma, sigma):
        """

        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        A : float
            Amplitude of the voigt.

        centre: float
            The position of the peak position

        gamma: float
            The full width half maximum of the gaussian.

        sigma: float
            The standard deviation of the lorentzian.

        Returns
        -------
        """
        super().__init__(specshape)

        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(0, np.inf)
        p1.sethasgradient(False)
        self._addparameter(p1)

        p2 = Parameter('centre', centre)
        p2.setboundaries(-np.inf, np.inf)
        p2.sethasgradient(False)
        self._addparameter(p2)

        p3 = Parameter('gamma', gamma)
        p3.setboundaries(self.dispersion/10, np.inf)
        p3.sethasgradient(False)
        self._addparameter(p3)

        p4 = Parameter('sigma', sigma)
        p4.setboundaries(self.dispersion/10, np.inf)
        p4.sethasgradient(False)
        self._addparameter(p4)

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        E = self.energy_axis
        self.data = self.voigt(E, self.parameters[0].getvalue(),
                               self.parameters[1].getvalue(),
                               self.parameters[2].getvalue(),
                               self.parameters[3].getvalue())

    def voigt(self, x, A, centre, gamma, sigma):
        """
        Voigt function.
        #todo add citation where to find definition of this calculation
        """
        z = ((x-centre) + 1j*gamma)/(np.sqrt(2)*sigma)
        # w_z = np.exp(-z**2)*special.erfc(-1j*z)
        w_z = special.wofz(z)
        V = np.real(w_z)/(sigma*np.sqrt(2*np.pi))
        V[np.isnan(V)] = 0
        V[np.isinf(V)] = 0
        V = V/V.max()
        return A*V
        # return A*np.pi**-1*0.5*dT/((x-x0)**2+(0.5*dT)**2)

    def get_fwhm(self):
        sigma = self.parameters[2].getvalue()
        fg = 2 * sigma * np.sqrt(2 * np.log(2))
        gamma = self.parameters[3].getvalue()
        fl = 2 * gamma
        fwhm = 0.5346 * fl + np.sqrt(0.166 * fl**2+fg**2)
        return fwhm

    def getgradient(self, parameter):
        return None

# todo add gradients



"""
copyright University of Antwerp 2021
author: Jo Verbeeck and Daen Jannis
"""
from pyEELSMODEL.core.component import Component
from pyEELSMODEL.core.parameter import Parameter
import numpy as np
import logging

logger = logging.getLogger(__name__)


class Exponential(Component):
    """
        Initialises an Exponential component

        Parameters
        ----------
        specshape : Spectrumshape
            The spectrum shape used to model
        A : float
            Amplitude of the exponential

        b: float
            The constant in the exponential

        Returns
        -------

    """

    def __init__(self, specshape, A, b):
        super().__init__(specshape)

        p1 = Parameter('A', A)
        p1.setlinear(True)
        p1.setboundaries(0, np.inf)
        p1.sethasgradient(True)
        self._addparameter(p1)

        p2 = Parameter('b', b)
        p2.setlinear(False)
        p2.sethasgradient(True)
        p2.setboundaries(np.inf, 0)

        self._addparameter(p2)

        self._setname('Exponential')

    def calculate(self):
        if self.suppress:
            self.data[:]=0
            self.setunchanged()
            return
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        if p1.ischanged() or p2.ischanged():
            A = p1.getvalue()
            b = p2.getvalue()
            self.data = self.exponential_function(A, b)
        self.setunchanged()  # put parameters to unchanged

    def exponential_function(self, A, b):
        # if sigma<self.dispersion: #sigma < dispersion makes no sense
        #    sigma=self.dispersion
        return A * np.exp(b * self.energy_axis)

    def autofit(self, spectrum, istart, istop):
        """
        Perform a linear least square fitting to get an estimate of the
        exponential fit.

        Parameters
        ----------
        spectrum: Spectrum
            The spectrum used to get the estimate from
        istart: uint
            The starting index over which to get the linear least square fit
        istop: uint
            The ending index over which to get the linear least square fit

        """
        if istart > istop:
            logger.warning(r'Start index is larger than stop index')
            return

        subdata = spectrum.data[istart:istop]
        boolean = subdata >= 1.
        ln_y = np.log(subdata[boolean])

        A = np.zeros((boolean.sum(), 2))
        A[:, 0] = np.ones(boolean.sum())
        A[:, 1] = self.energy_axis[istart:istop][boolean]
        coeff, resi, rank, sing = np.linalg.lstsq(A, ln_y, rcond=-1)
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        p1.setvalue(np.exp(coeff[0]))
        p2.setvalue(coeff[1])
        self.calculate()

    def getgradient(self, parameter):
        p1 = self.parameters[0]
        p2 = self.parameters[1]
        A = p1.getvalue()
        b = p2.getvalue()
        en = self.energy_axis

        if parameter == p1:
            return self.exponential_function(1, b)
        if parameter == p2:
            return en * self.exponential_function(A, b)

        return None



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  9 20:46:09 2021

@author: joverbee
"""
import numpy as np
import logging
from pyEELSMODEL.operators.deconvolutions.deconvolution import Deconvolution
from pyEELSMODEL.core.multispectrum import MultiSpectrum
from pyEELSMODEL.core.spectrum import Spectrum
from tqdm import tqdm

logger = logging.getLogger(__name__)


class RichardsonLucyDeconvolution(Deconvolution):
    """
    Richardson lucy deconvolution as described in
    10.1016/j.ultramic.2009.06.010.

    """
    def __init__(self, spectrum, llspectrum, iterations=5):
        """
        Parameters
        ----------
        spectrum: Spectrum or MultiSpectrum
            The spectrum which needs to be deconvolved. Best practice is to
            remove the background before deconvolving.
        llspectrum: Spectrum or MultiSpectrum
            The low loss which is used for the deconvolution
        iterations: uint
            The number of iterations used in the Richardson Lucy deconvolution
            (default: 5).

        """
        super().__init__(spectrum, llspectrum)
        self.iterations = iterations
        self.restored = None

    @property
    def iterations(self):
        return self._iterations

    @iterations.setter
    def iterations(self, n_iter):
        self._iterations = n_iter

    def richardsonlucy(self, signal, psf, iterations):
        """
        RichardsonLucy algorithm
        """
        imax = psf.argmax()
        result = np.array(signal).copy()
        mimax = psf.size - 1 - imax

        for i in range(iterations):
            first = np.convolve(psf, result)[imax: imax + psf.size]
            result *= np.convolve(psf[::-1], signal /
                                  first)[mimax:mimax + psf.size]

            # plt.figure()
            # plt.plot(result)
            # plt.title(i)

        return result

    # def ISRA(self, signal, psf, iterations):
    #     imax = psf.argmax()
    #     result = np.array(signal).copy()
    #     mimax = psf.size - 1 - imax
    #     mode = 'full'
    #     for i in range(iterations):
    #         # first = np.convolve(psf, result, mode=mode)
    #         first = result.copy()
    #         teller =  np.convolve(signal, psf[::-1], mode=mode)
    #         noemer =  np.convolve(np.convolve(psf, first, mode=mode),
    #         psf[::-1], mode=mode)
    #         # update = teller/noemer
    #         # print(update.sum())
    #         # result *= first*update
    #         # print(result.sum())
    #
    #         # plt.figure()
    #         # plt.plot(result)
    #     return result

    def multi_deconvolve(self):
        shape = (self.spectrum.xsize, self.spectrum.ysize)
        ms = self.spectrum.copy()

        for index in tqdm(np.ndindex(shape)):
            islice = np.s_[index]
            self.spectrum.setcurrentspectrum(index)
            self.llspectrum.setcurrentspectrum(index)
            restore = self.richardsonlucy(self.spectrum.data,
                                          self.llspectrum.data,
                                          self.iterations)
            ms.multidata[islice] = restore

        return ms

    def deconvolve(self):
        # psf = self.llspectrum.data / self.llspectrum.data.sum()
        # signal = self.spectrum.data / self.spectrum.data.sum()
        # restore = self.ISRA(signal, psf, self.iterations)

        if type(self.spectrum) is MultiSpectrum:
            s = self.multi_deconvolve()
            self.restored = s
        else:
            restore = self.richardsonlucy(self.spectrum.data,
                                          self.llspectrum.data,
                                          self.iterations)
            s = Spectrum(self.spectrum.get_spectrumshape(), data=restore)
            self.restored = s
        return s



import sys

import pytest

from pyEELSMODEL.components.powerlaw import PowerLaw
from pyEELSMODEL.components.exponential import Exponential
from pyEELSMODEL.components.fast_background import FastBG2
from pyEELSMODEL.components.polynomial import Polynomial

from pyEELSMODEL.core.spectrum import Spectrumshape, Spectrum
from pyEELSMODEL.core.multispectrum import MultiSpectrum, MultiSpectrumshape
from pyEELSMODEL.operators.backgroundremoval import BackgroundRemoval
import numpy as np
from pyEELSMODEL.fitters.lsqfitter import LSQFitter
from pyEELSMODEL.fitters.linear_fitter import LinearFitter
from pyEELSMODEL.fitters.minimizefitter import MinimizeFitter
sys.path.append("..")  # Adds higher directory to python modules path.


def make_multispectrum():
    specshape = Spectrumshape(1, 100, 1024)
    shape = (10, 10)
    A_ar = np.reshape(np.linspace(1000, 10000, shape[0] * shape[1]), shape)
    r_ar = np.reshape(np.linspace(2, 5, shape[0] * shape[1]), shape)
    data = np.empty((shape[0], shape[1], specshape.size))
    for index in (np.ndindex(shape)):
        islice = np.s_[index]
        pw = PowerLaw(specshape, A=A_ar[islice], r=r_ar[islice])
        pw.calculate()
        data[islice] = pw.data

    mshape = MultiSpectrumshape(specshape.dispersion, specshape.offset,
                                specshape.size, shape[0], shape[1])
    s_mult = MultiSpectrum(mshape, data=data)
    return s_mult


def test_init():
    specshape = Spectrumshape(1, 10, 100)
    s = Spectrum(specshape)

    with pytest.raises(ValueError):
        BackgroundRemoval(s, signal_range=(5., 20.))
    with pytest.raises(ValueError):
        BackgroundRemoval(s, signal_range=(5., 120.))
    with pytest.raises(ValueError):
        BackgroundRemoval(s, signal_range=(50., 120.))
    with pytest.raises(ValueError):
        BackgroundRemoval(s, signal_range=((50., 60.), (20., 30.)))
    with pytest.raises(ValueError):
        BackgroundRemoval(s, signal_range=((70., 60.), (20., 30.)))
    with pytest.raises(ValueError):
        BackgroundRemoval(s, signal_range=(50., 60.), model_type='Hello')
    with pytest.raises(ValueError):
        BackgroundRemoval(s, signal_range=(50., 60.),
                          non_linear_fitter='Hello')
    with pytest.raises(ValueError):
        BackgroundRemoval(s, signal_range=(50., 60.), linear_fitting=True)
    with pytest.raises(ValueError):
        BackgroundRemoval(s, signal_range=(50., 60.), order=-5)
    with pytest.raises(TypeError):
        BackgroundRemoval(s, signal_range=(50., 60.), order=2.1)

    back = BackgroundRemoval(s, signal_range=((20., 30.), (40., 50.)))
    assert back.two_area is True


def test_set_indices():
    specshape = Spectrumshape(1, 10, 100)
    s = Spectrum(specshape)
    back = BackgroundRemoval(s, signal_range=(50., 60.))
    assert back.indices[0] == 40
    assert back.indices[1] == 50

    back = BackgroundRemoval(s, signal_range=((20., 30.), (50., 60.)))
    assert back.indices[0][0] == 10
    assert back.indices[0][1] == 20
    assert back.indices[1][0] == 40
    assert back.indices[1][1] == 50


def test_include_areas():
    specshape = Spectrumshape(1, 10, 100)
    s = Spectrum(specshape)
    back = BackgroundRemoval(s, signal_range=(50., 60.))
    back.include_areas()
    test = np.ones(s.size, dtype=bool)
    test[40:50] = False
    comparison = test == s.exclude
    assert comparison.all()

    back = BackgroundRemoval(s, signal_range=((20., 30.), (50., 60.)))
    back.include_areas()
    test = np.ones(s.size, dtype=bool)
    test[40:50] = False
    test[10:20] = False
    comparison = test == s.exclude
    assert comparison.all()


def test_make_model_type():
    specshape = Spectrumshape(1, 10, 100)
    s = Spectrum(specshape, data=np.ones(specshape.size))

    models = ['Polynomial', 'Exponential', 'Powerlaw', 'FastBG']
    comp_list = [Polynomial, Exponential, PowerLaw, FastBG2]
    for index, model_type in enumerate(models):
        back = BackgroundRemoval(s, signal_range=(50., 60.),
                                 model_type=model_type)
        back.make_background_model()
        assert type(back.model.components[0]) is comp_list[index]


def test_set_fit_type():
    specshape = Spectrumshape(1, 10, 100)
    s = Spectrum(specshape, data=np.ones(specshape.size))
    back = BackgroundRemoval(s, signal_range=(50., 60.), model_type='FastBG',
                             linear_fitting=True)
    back.make_background_model()
    back.set_fit_type()
    assert type(back.fitter) is LinearFitter

    back = BackgroundRemoval(s, signal_range=(50., 60.))
    back.make_background_model()
    back.set_fit_type()
    assert type(back.fitter) is LSQFitter

    back = BackgroundRemoval(s, signal_range=(50., 60.),
                             non_linear_fitter='ML')
    back.make_background_model()
    back.set_fit_type()
    assert type(back.fitter) is MinimizeFitter


def test_reset_exclude_after_fit():
    specshape = Spectrumshape(1, 10, 100)
    s = Spectrum(specshape, data=np.ones(specshape.size))
    back = BackgroundRemoval(s, signal_range=(50., 60.))
    back.calculate()

    assert not s.exclude.all()


def test_background_subtraction():
    specshape = Spectrumshape(1, 175, 2048)
    fbg = PowerLaw(specshape, A=10000, r=3.)
    fbg.calculate()
    s = Spectrum(specshape, data=fbg.data)
    back = BackgroundRemoval(s, signal_range=(175., 275.))
    result = back.calculate()
    assert np.allclose(result.data, np.zeros(s.size), atol=1e-3)


def test_fast_fit():
    s = make_multispectrum()
    back = BackgroundRemoval(s, signal_range=(100., 200.))
    fast_back = back.fast_calculate_multi()
    check = np.zeros((s.xsize, s.ysize, s.size))
    assert np.allclose(fast_back.multidata, check, atol=1e-3)


def test_model_fit():
    s = make_multispectrum()
    back = BackgroundRemoval(s, signal_range=(100., 200.))
    s_back = back.calculate_multi()
    check = np.zeros((s.xsize, s.ysize, s.size))
    assert np.allclose(s_back.multidata, check, atol=1e-3)


def main():
    test_init()
    test_set_indices()
    test_include_areas()
    test_make_model_type()
    test_set_fit_type()
    test_reset_exclude_after_fit()
    test_background_subtraction()
    test_fast_fit()
    test_model_fit()


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
from quadprog import solve_qp
from pyEELSMODEL.components.CLedge.coreloss_edge import CoreLossEdge
from pyEELSMODEL.components.fixedpattern import FixedPattern
from pyEELSMODEL.components.linear_background import LinearBG
from pyEELSMODEL.components.constrained_gdoslin import ConstrainedGDOSLin
from pyEELSMODEL.fitters.linear_fitter import LinearFitter
from pyEELSMODEL.components.powerlaw import PowerLaw


class QuadraticFitter(LinearFitter):
    '''
    A linear fitter class which uses as input a Spectrum or Multispectrum and
    Model.
    '''

    def __init__(self, spectrum, model, use_weights=True):

        '''
        Initialises a QuadraticFitter instance.

        Parameters
        ----------
        spectrum : Spectrum or Multispectrum
            The experimental data used which needs to be fitted
        model : Model
            The model used to fit the experimental data. Model should only
            contain linear parameters. The background should also contain the
            right background parameters

        Returns
        -------
        An instance of a QuadraticFitter.

        '''
        # model can only be check if it is add
        if not model.islinear():
            raise ValueError(r'There are non-linear parameters in the model')

        super().__init__(spectrum, model, use_weights=use_weights)

        self.G_matrix = None
        self.h_vector = None
        self.eq_idx = []
        self.use_constrain_coreloss = True  # chose to let the coreloss model
        # to be positive constrained

    @property
    def G_matrix(self):
        return self._G_matrix

    @G_matrix.setter
    def G_matrix(self, G):
        self._G_matrix = G

    @property
    def h_vector(self):
        return self._h_vector

    @h_vector.setter
    def h_vector(self, h):
        self._h_vector = h

    def _linearfit(self, x, y):
        """
        Estimates the linear fit from x and y

        Args:
            x (_type_): _description_
            y (_type_): _description_

        Returns:
            _type_: _description_
        """
        n = x.size
        noemer_a = n*np.sum(x**2) - np.sum(x)**2

        a = (n*np.sum(x*y) - np.sum(x)*np.sum(y))/noemer_a
        b = 1/n*(np.sum(y) - a*np.sum(x))

        return a*x+b

    def calculate_weights(self, eps=1.):
        """
        Small function which estimates the weigths used in the quadratic
        programming. It tries to fit a powerlaw to the spectrum and uses this
        as input. If that does not work, a linear fit is performed which used
        this result as input for the weights.

        Args:
            eps (float):Error terms added to weigths. Defaults to 1..

        Returns:
            weigths (np.array): Weights used in the quadratic programming
            describing poisson process
        """
        powlaw_bg = PowerLaw(self.spectrum.get_spectrumshape(), 1., 3.)
        # Pass dummies: A = 1. and r = 3.
        check = powlaw_bg.autofit(self.spectrum, 0, self.spectrum.size)
        ndata = powlaw_bg.data

        if not check:
            # do linear fit on data
            x = self.spectrum.energy_axis
            y = self.spectrum.data
            ndata = self._linearfit(x, y)
            ndata[ndata < 0] = 0

        # powlaw_bg.autofit_twowindow(self.spectrum, 0, self.spectrum.size)
        # Quick and dirty fit. ~20% faster overall.

        msk_include = np.logical_not(self.spectrum.exclude)
        weights = ndata[msk_include]

        return 1. / (weights + eps)

    def perform_fit(self):
        Am = self.A_matrix
        if Am is None or self.model.ischanged() or not self.same_A_matrix:
            self.calculate_A_matrix()  # makes it slow
            self.model.setchanged(False)
            self.get_quadratic_bounds()

        if self.model.hasconvolutor():
            A = self.convolute_A_matrix()[np.invert(self.spectrum.exclude), :]
        else:
            A = self.A_matrix[np.invert(self.spectrum.exclude), :]

        self.calculate_y_vector()
        y = self.y_vector

        if self.use_weights:
            weights = self.calculate_weights()
            tmp = np.sqrt(weights)
            AW = A * tmp[:, np.newaxis]
            yW = y * tmp
        else:
            AW = A
            yW = y

        G = self.G_matrix
        h = self.h_vector

        P = np.dot(np.transpose(AW), AW)
        q = -np.dot(np.transpose(AW), yW)
        result = solve_qp(P, -q, G.T, h, self.number_equality)[0]

        self.coeff = result

    def get_quadratic_bounds(self):
        """
        another method to get the boundaries
        :return:
        """
        if self.A_matrix is None:
            raise ValueError(r'A matrix should first be calculated before G'
                             r' can be determined')

        # following the convention of the quadratic programming, the equality
        # constrains should
        # be placed before the inequality ones

        inequality_arrays = []  # array stores the inequality conditions
        equality_arrays = []  # array stores the equality conditions

        used_comp = []  # list to make sure that the same component with
        # different
        # parameters does not contribute redundant rows to the constrains
        for param in (self.model.getfreelinparameters()):
            comp = self.model.getcomponentbyparameter(param)

            already_used = comp in used_comp
            is_bg = (type(comp) is LinearBG) & (not already_used)
            is_constrained_gdoslin = (type(comp) is ConstrainedGDOSLin) & (not already_used)

            is_coreloss = (isinstance(comp, CoreLossEdge)) & self.use_constrain_coreloss
            is_fixed = (isinstance(comp, FixedPattern)) & self.use_constrain_coreloss

            if is_coreloss or is_fixed:
                cst_array = np.zeros((1, self.A_matrix.shape[1]))
                indey = self.get_param_index(param)
                cst_array[:, indey] = 1
                used_comp.append(comp)
                inequality_arrays.append(cst_array)

            elif is_bg:
                indey = self.get_param_index(param)
                bgG = comp.get_convex_boundaries()
                # if no constrains are given it should skip the rest
                if bgG is None:
                    continue
                cst_array = np.zeros((bgG.shape[0], self.A_matrix.shape[1]))
                # cst_array[:, indey:indey+bgG.shape[1]] = bgG
                # used_comp.append(comp)
                inequality_arrays.append(cst_array)

            elif is_constrained_gdoslin:
                indey = self.get_param_index(param)

                g_mtrx = comp.get_equality_constraints()
                cst_array = np.zeros((g_mtrx.shape[0], self.A_matrix.shape[1]))
                cst_array[:, indey:(indey + g_mtrx.shape[1])] = g_mtrx
                used_comp.append(comp)
                equality_arrays.append(cst_array)

        if len(equality_arrays) == 0:
            self.G_matrix = np.concatenate(inequality_arrays, axis=0)
            self.number_equality = 0

        elif len(inequality_arrays) == 0:
            self.G_matrix = np.concatenate(equality_arrays, axis=0)
            self.number_equality = self.G_matrix.shape[0]

        else:
            eq = np.concatenate(equality_arrays, axis=0)
            ineq = np.concatenate(inequality_arrays, axis=0)
            self.G_matrix = np.concatenate((eq, ineq), axis=0)
            self.number_equality = eq.shape[0]

        self.h_vector = np.zeros(self.G_matrix.shape[0])

    def plot_conv(self):
        """
        Shows the convoluted components of the model.
        """
        conv_A_matrix = self.convolute_A_matrix()
        self.set_fit_values()
        indices = [0]
        names = []
        for comp in self.model.components:
            indices.append(len(comp.getfreelinparameters())+indices[-1])
            names.append(comp.name)

        print(indices)

        fig, ax = plt.subplots()
        Eax = self.spectrum.energy_axis
        ax.plot(Eax, self.spectrum.data, color='black',
                label='Experimental data')
        for ii in range(len(indices)-1):
            prod = conv_A_matrix[:, indices[ii]:indices[ii+1]]\
                   * self.coeff[indices[ii]:indices[ii+1]]
            ax.plot(Eax, prod.sum(1), label=names[ii])
        ax.plot(Eax, self.model.data, label='Model')
        ax.legend()
